from typing import cast

import numpy as np
import numpy.typing as npt


def bits_to_int(input: npt.ArrayLike) -> int | npt.NDArray[np.integer]:
    r"""
    Converts a bit array to its integer representation (LSB first).

    Parameters:
        input: An $N$-dimensional array of $0$s and $1$s. The least significant bit (LSB) is the first element in the last dimension.

    Returns:
        An integer or an $(N-1)$-dimensional array of integers. The last dimension of the input is collapsed into an integer representation while all preceding dimensions are preserved.

    Examples:
        >>> komm.bits_to_int([0, 0, 0, 0, 1])
        16

        >>> komm.bits_to_int([0, 1, 0, 1, 1])
        26

        >>> komm.bits_to_int([0, 1, 0, 1, 1, 0, 0, 0])
        26

        >>> komm.bits_to_int([[0, 0], [1, 0], [0, 1], [1, 1]])  # Each row is independently converted to an integer
        array([0, 1, 2, 3])
    """
    if np.ndim(input) > 1:
        return np.apply_along_axis(bits_to_int, -1, input)
    input = np.asarray(input, dtype=np.uint8)
    packed = np.packbits(input, bitorder="little")
    return int.from_bytes(packed, byteorder="little")


def int_to_bits(input: npt.ArrayLike, width: int) -> npt.NDArray[np.integer]:
    r"""
    Converts an integer, or array of integers, to their bit representations (LSB first).

    Parameters:
        input: An integer or an $N$-dimensional array of integers.

        width: The width of the bit representation.

    Returns:
        An $(N+1)$-dimensional array of $0$s and $1$s, where the last dimension contains the bit representation of the input, with the least significant bit (LSB) as the first element.

    Examples:
        >>> komm.int_to_bits(16, width=5)
        array([0, 0, 0, 0, 1])

        >>> komm.int_to_bits(26, width=5)
        array([0, 1, 0, 1, 1])

        >>> komm.int_to_bits(26, width=8)
        array([0, 1, 0, 1, 1, 0, 0, 0])

        >>> komm.int_to_bits([0, 1, 2, 3], width=2)
        array([[0, 0],
               [1, 0],
               [0, 1],
               [1, 1]])

        >>> komm.int_to_bits([0, 1, 2, 3], width=4)
        array([[0, 0, 0, 0],
               [1, 0, 0, 0],
               [0, 1, 0, 0],
               [1, 1, 0, 0]])

        >>> komm.int_to_bits([[0, 1], [2, 3]], width=2)
        array([[[0, 0],
                [1, 0]],
        <BLANKLINE>
               [[0, 1],
                [1, 1]]])
    """
    if np.ndim(input) > 0:
        input = np.asarray(input, dtype=object)
        bits = np.array([int_to_bits(x, width) for x in input.flat])
        return bits.reshape(input.shape + (width,))
    input = cast(int, input)
    return np.array([(input >> i) & 1 for i in range(width)])



from abc import ABC, abstractmethod
from functools import cached_property

import numpy as np
import numpy.typing as npt


class ScalarQuantizer(ABC):
    @cached_property
    @abstractmethod
    def levels(self) -> npt.NDArray[np.floating]:
        r"""
        The quantizer levels $v_0, v_1, \ldots, v_{L-1}$.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def thresholds(self) -> npt.NDArray[np.floating]:
        r"""
        The quantizer finite thresholds $t_1, t_2, \ldots, t_{L-1}$.
        """
        raise NotImplementedError

    @abstractmethod
    def digitize(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Returns the quantization indices for the input signal.

        Parameters:
            input: The input signal $x$ to be digitized.

        Returns:
            output: The integer indices of the quantization levels for each input sample.
        """
        tiled = np.tile(input, reps=(self.thresholds.size, 1)).transpose()
        output = np.sum(tiled >= self.thresholds, axis=1)
        return output

    @abstractmethod
    def quantize(self, input: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Quantizes the input signal.

        Parameters:
            input: The input signal $x$ to be quantized.

        Returns:
            output: The quantized signal $y$.
        """
        return self.levels[self.digitize(input)]



import math
from functools import partial
from typing import Literal

import numpy as np
import numpy.typing as npt

LogBase = float | Literal["e"]


def assert_is_log_base(value: float | str) -> None:
    if (isinstance(value, str) and value != "e") or (
        isinstance(value, float) and value <= 0.0
    ):
        raise ValueError("log base must be 'e' or a positive float")


def assert_is_probability(value: float) -> None:
    if not 0 <= value <= 1:
        raise ValueError("probability must be between 0 and 1")


class PMF(npt.NDArray[np.floating]):
    def __new__(cls, values: npt.ArrayLike):
        arr = np.asarray(values, dtype=float)
        if arr.ndim != 1:
            raise ValueError("PMF must be a 1D array")
        if not np.all(arr >= 0.0):
            raise ValueError("PMF must be non-negative")
        if not np.isclose(arr.sum(), 1.0):
            raise ValueError("PMF must sum to 1.0")
        obj = arr.view(cls)
        return obj

    def __array_finalize__(self, obj: npt.NDArray[np.floating] | None) -> None:
        if obj is None:
            return


def random_pmf(size: int) -> PMF:
    pmf = np.random.rand(size)
    return PMF(pmf / pmf.sum())


class TransitionMatrix(npt.NDArray[np.floating]):
    def __new__(cls, values: npt.ArrayLike):
        arr = np.asarray(values, dtype=float)
        if arr.ndim != 2:
            raise ValueError("Transition matrix must be a 2D array")
        if not np.all(arr >= 0.0):
            raise ValueError("Transition matrix must be non-negative")
        if not np.allclose(arr.sum(axis=1), 1.0):
            raise ValueError("Rows of transition matrix must sum to 1.0")
        obj = arr.view(cls)
        return obj

    def __array_finalize__(self, obj: npt.NDArray[np.floating] | None) -> None:
        if obj is None:
            return


def entropy(pmf: npt.ArrayLike, base: LogBase = 2.0) -> float:
    r"""
    Computes the entropy of a random variable with a given pmf. Let $X$ be a random variable with pmf $p_X$ and alphabet $\mathcal{X}$. Its entropy is given by
    $$
        \mathrm{H}(X) = \sum_{x \in \mathcal{X}} p_X(x) \log \frac{1}{p_X(x)}.
    $$
    By default, the base of the logarithm is $2$, in which case the entropy is measured in bits. For more details, see <cite>CT06, Sec. 2.1</cite>.

    Parameters:
        pmf: The probability mass function $p_X$ of the random variable. It must be a valid pmf, that is, all of its values must be non-negative and sum up to $1$.

        base: The base of the logarithm to be used. It must be a positive float or the string `'e'`. The default value is `2.0`.

    Returns:
        The entropy $\mathrm{H}(X)$ of the random variable.

    Examples:
        >>> komm.entropy([1/4, 1/4, 1/4, 1/4])  # doctest: +FLOAT_CMP
        np.float64(2.0)

        >>> komm.entropy(pmf=[1/3, 1/3, 1/3], base=3.0)  # doctest: +FLOAT_CMP
        np.float64(1.0)

        >>> komm.entropy([0.5, 0.5], base='e')  # doctest: +FLOAT_CMP
        np.float64(0.6931471805599453)
    """
    pmf = PMF(pmf)
    assert_is_log_base(base)
    if base == "e":
        return -np.dot(pmf, np.log(pmf, where=(pmf > 0)))
    elif base == 2.0:
        return -np.dot(pmf, np.log2(pmf, where=(pmf > 0)))
    else:
        return -np.dot(pmf, np.log(pmf, where=(pmf > 0))) / np.log(base)


def binary_entropy(p: float) -> float:
    r"""
    Computes the binary entropy function. For a given probability $p$, it is defined as
    $$
        \Hb(p) = p \log_2 \frac{1}{p} + (1 - p) \log_2 \frac{1}{1 - p},
    $$
    and corresponds to the [entropy](/ref/entropy) of a Bernoulli random variable with parameter $p$.

    Parameters:
        p: A probability value. It must satisfy $0 \leq p \leq 1$.

    Returns:
        The value of the binary entropy function $\Hb(p)$.

    Examples:
        >>> [komm.binary_entropy(p) for p in [0.0, 0.25, 0.5, 0.75, 1.0]]  # doctest: +FLOAT_CMP
        [0.0, 0.8112781244591328, 1.0, 0.8112781244591328, 0.0]
    """
    assert_is_probability(p)
    if p in {0.0, 1.0}:
        return 0.0
    return -p * math.log2(p) - (1 - p) * math.log2(1 - p)


def relative_entropy(
    pmf: npt.ArrayLike,
    qmf: npt.ArrayLike,
    base: LogBase = 2.0,
) -> float:
    r"""
    Computes the relative entropy (Kullback–Leibler divergence) between two <span>pmf</span>s. Let $p$ and $q$ be two <span>pmf</span>s over the same alphabet $\mathcal{X}$. The relative entropy of $p$ with respect to $q$ is defined as
    $$
        \mathrm{D}(p \parallel q) = \sum_{x \in \mathcal{X}} p(x) \log \frac{p(x)}{q(x)}.
    $$
    Note that, in general, $\mathrm{D}(p \parallel q) \neq \mathrm{D}(q \parallel p)$. For more details, see <cite>CT06, Sec. 2.3</cite>.

    Parameters:
        pmf: The probability mass function $p$. It must be a valid pmf, that is, all of its values must be non-negative and sum up to $1$.

        qmf: The probability mass function $q$. It must be a valid pmf, that is, all of its values must be non-negative and sum up to $1$.

        base: The base of the logarithm to be used. It must be a positive float or the string `'e'`. The default value is `2.0`.

    Returns:
        The relative entropy $\mathrm{D}(p \parallel q)$ between the two <span>pmf</span>s.

    Examples:
        >>> komm.relative_entropy([1/2, 1/2], [1/2, 1/2])  # doctest: +FLOAT_CMP
        np.float64(0.0)

        >>> komm.relative_entropy([1/2, 1/2], [3/4, 1/4])  # doctest: +FLOAT_CMP
        np.float64(0.20751874963942185)

        >>> komm.relative_entropy([3/4, 1/4], [1/2, 1/2])  # doctest: +FLOAT_CMP
        np.float64(0.18872187554086717)

        >>> komm.relative_entropy([1/2, 1/2], [0, 1])  # doctest: +FLOAT_CMP
        np.float64(inf)
    """
    pmf = PMF(pmf)
    qmf = PMF(qmf)
    assert_is_log_base(base)
    with np.errstate(divide="ignore"):
        if base == "e":
            return np.dot(pmf, np.log(pmf / qmf))
        elif base == 2.0:
            return np.dot(pmf, np.log2(pmf / qmf))
        else:
            return np.dot(pmf, np.log(pmf / qmf)) / np.log(base)


def mutual_information(
    input_pmf: npt.ArrayLike,
    transition_matrix: npt.ArrayLike,
    base: LogBase = 2.0,
) -> float:
    input_pmf = PMF(input_pmf)
    assert_is_log_base(base)
    transition_matrix = TransitionMatrix(transition_matrix)
    output_pmf = np.dot(input_pmf, transition_matrix)
    entropy_output_prior = entropy(output_pmf, base=base)
    entropy_output_posterior = np.dot(
        input_pmf,
        np.apply_along_axis(
            func1d=partial(entropy, base=base),
            axis=1,
            arr=transition_matrix,
        ),
    )
    return entropy_output_prior - entropy_output_posterior


def arimoto_blahut(
    transition_matrix: npt.NDArray[np.floating],
    initial_guess: npt.NDArray[np.floating],
    max_iter: int,
    tol: float,
) -> npt.NDArray[np.floating]:
    r"""
    Arimoto–Blahut algorithm for channel capacity. See <cite>CT06, Sec. 10.8</cite>.
    """
    p = transition_matrix
    r = initial_guess
    last_r = np.full_like(r, fill_value=np.inf)
    iters = 0
    while iters < max_iter and np.amax(np.abs(r - last_r)) > tol:
        last_r = r
        q = r[np.newaxis].T * p
        qsum = np.sum(q, axis=0)
        zero_indices = qsum == 0
        qsum_copy = qsum.copy()
        qsum_copy[zero_indices] = 1
        q /= qsum_copy
        q[:, zero_indices] = 0
        r = np.prod(q**p, axis=1)
        r /= np.sum(r, axis=0)
        iters += 1
    return r



from .BarkerSequence import BarkerSequence
from .BinarySequence import BinarySequence
from .GoldSequence import GoldSequence
from .KasamiSequence import KasamiSequence
from .LFSRSequence import LFSRSequence
from .WalshHadamardSequence import WalshHadamardSequence

__all__ = [
    "BarkerSequence",
    "BinarySequence",
    "GoldSequence",
    "KasamiSequence",
    "LFSRSequence",
    "WalshHadamardSequence",
]



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from . import base


@dataclass
class RaisedCosinePulse(base.Pulse):
    r"""
    Raised cosine pulse. It is a [pulse](/ref/Pulse) with waveform given by
    $$
        h(t) = \operatorname{sinc}(t) \frac{\cos(\pi \alpha t)}{1 - (2 \alpha t)^2},
    $$
    where $\alpha$ is the *roll-off factor*, which must satisfy $0 \leq \alpha \leq 1$. Its spectrum is given by
    $$
        \hat{h}(f) = \begin{cases}
            1, & |f| \leq f_1, \\\\[1ex]
            \dfrac{1}{2} \left( 1 + \cos \left( \pi \dfrac{|f| - f_1}{f_2 - f_1}\right) \right), & f_1 \leq |f| \leq f_2, \\\\[1ex]
            0, & \text{otherwise}.
        \end{cases}
    $$
    where $f_1 = (1 - \alpha) / 2$ and $f_2 = (1 + \alpha) / 2$.

    The raised cosine pulse is depicted below for $\alpha = 0.25$, and for $\alpha = 0.75$.

    <div class="centered" markdown>
      <span>
        ![Raised cosine pulse with roll-off factor 0.25.](/figures/pulse_raised_cosine_25.svg)
      </span>
      <span>
        ![Raised cosine pulse with roll-off factor 0.75.](/figures/pulse_raised_cosine_75.svg)
      </span>
    </div>

    For more details, see <cite>PS08, Sec. 9.2-1</cite>.

    Notes:
        - For $\alpha = 0$ it reduces to the [sinc pulse](/ref/SincPulse).
        - For $\alpha = 1$ it becomes what is known as the _full cosine roll-off pulse_.

    Attributes:
        rolloff: The roll-off factor $\alpha$ of the pulse. Must satisfy $0 \leq \alpha \leq 1$. The default value is `1.0`.

    Examples:
        >>> pulse = komm.RaisedCosinePulse(rolloff=0.25)
        >>> pulse.waveform([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75]).round(4)
        array([0.2904, 0.6274, 0.897 , 1.    , 0.897 , 0.6274, 0.2904])
        >>> pulse.spectrum([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75])
        array([0. , 0.5, 1. , 1. , 1. , 0.5, 0. ])
    """

    rolloff: float = 1.0

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        a = self.rolloff
        t = np.asarray(t) + 1e-8  # TODO: Improve this workaround
        return np.sinc(t) * np.cos(np.pi * a * t) / (1 - (2 * a * t) ** 2)

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        a = self.rolloff
        f = np.asarray(f)
        if a == 0:
            return 1.0 * (abs(f) < 0.5)
        f1 = (1 - a) / 2
        f2 = (1 + a) / 2
        band1 = abs(f) < f1
        band2 = (f1 < abs(f)) * (abs(f) < f2)
        return 1.0 * band1 + (0.5 * (1 + np.cos(np.pi * (abs(f) - f1) / a))) * band2

    @cached_property
    def support(self) -> tuple[float, float]:
        return (-np.inf, np.inf)



from typing_extensions import Self


class Integer:
    def __init__(self, value: int) -> None:
        self.value = value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.value})"

    def __str__(self) -> str:
        return str(self.value)

    @property
    def ambient(self) -> "Integers":
        return Integers()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.value == other.value

    def __add__(self, other: Self) -> Self:
        return self.__class__(self.value + other.value)

    def __sub__(self, other: Self) -> Self:
        return self.__class__(self.value - other.value)

    def __neg__(self) -> Self:
        return self.__class__(-self.value)

    def __mul__(self, other: Self) -> Self:
        return self.__class__(self.value * other.value)

    def __rmul__(self, other: int) -> Self:
        return self.__class__(self.value * other)

    def __divmod__(self, other: Self) -> tuple[Self, Self]:
        quotient, remainder = divmod(self.value, other.value)
        return self.__class__(quotient), self.__class__(remainder)

    def __floordiv__(self, other: Self) -> Self:
        return self.__class__(self.value // other.value)

    def __mod__(self, other: Self) -> Self:
        return self.__class__(self.value % other.value)


class Integers:
    def __call__(self, value: int) -> Integer:
        return Integer(value)

    @property
    def zero(self) -> Integer:
        return Integer(0)

    @property
    def one(self) -> Integer:
        return Integer(1)


def prime_factors(n: int) -> list[int]:
    factors: list[int] = []
    i = 2
    while i * i <= n:
        if n % i:
            i += 1
        else:
            n //= i
            factors.append(i)
    if n > 1:
        factors.append(n)
    return factors



from collections.abc import Callable, Sequence
from functools import cached_property
from typing import Any, TypedDict, TypeVar

import numpy as np
import numpy.typing as npt


class MetricMemory(TypedDict):
    paths: npt.NDArray[np.integer]
    metrics: npt.NDArray[np.floating]


Z = TypeVar("Z")
MetricFunction = Callable[[int, Z], float]


class FiniteStateMachine:
    r"""
    Finite-state machine (Mealy machine). It is defined by a *set of states* $\mathcal{S}$, an *input alphabet* $\mathcal{X}$, an *output alphabet* $\mathcal{Y}$, and a *transition function* $T : \mathcal{S} \times \mathcal{X} \to \mathcal{S} \times \mathcal{Y}$. Here, for simplicity, the set of states, the input alphabet, and the output alphabet are always taken as $\mathcal{S} = \\{ 0, 1, \ldots, |\mathcal{S}| - 1 \\}$, $\mathcal{X} = \\{ 0, 1, \ldots, |\mathcal{X}| - 1 \\}$, and $\mathcal{Y} = \\{ 0, 1, \ldots, |\mathcal{Y}| - 1 \\}$, respectively.

    For example, consider the finite-state machine whose state diagram depicted in the figure below.

    <figure markdown>
      ![Finite-state machine (Mealy machine) example.](/figures/mealy.svg)
    </figure>

    It has set of states $\mathcal{S} = \\{ 0, 1, 2, 3 \\}$, input alphabet $\mathcal{X} = \\{ 0, 1 \\}$, output alphabet $\mathcal{Y} = \\{ 0, 1, 2, 3 \\}$, and transition function $T$ given by the table below.

    | State | Input | Next state | Output |
    | :---: | :---: | :--------: | :----: |
    | $0$   | $0$   | $0$        | $0$    |
    | $0$   | $1$   | $1$        | $3$    |
    | $1$   | $0$   | $2$        | $1$    |
    | $1$   | $1$   | $3$        | $2$    |
    | $2$   | $0$   | $0$        | $3$    |
    | $2$   | $1$   | $1$        | $0$    |
    | $3$   | $0$   | $2$        | $2$    |
    | $3$   | $1$   | $3$        | $1$    |

    Parameters:
        next_states: The matrix of next states of the machine, of shape $|\mathcal{S}| \times |\mathcal{X}|$. The element in row $s$ and column $x$ should be the next state of the machine (an element in $\mathcal{S}$), given that the current state is $s \in \mathcal{S}$ and the input is $x \in \mathcal{X}$.

        outputs: The matrix of outputs of the machine, of shape $|\mathcal{S}| \times |\mathcal{X}|$. The element in row $s$ and column $x$ should be the output of the machine (an element in $\mathcal{Y}$), given that the current state is $s \in \mathcal{S}$ and the input is $x \in \mathcal{X}$.

    Examples:
        >>> fsm = komm.FiniteStateMachine(next_states=[[0,1], [2,3], [0,1], [2,3]], outputs=[[0,3], [1,2], [3,0], [2,1]])
    """

    def __init__(self, next_states: npt.ArrayLike, outputs: npt.ArrayLike):
        self.next_states = np.asarray(next_states)
        self.outputs = np.asarray(outputs)
        self._input_edges = np.full((self.num_states, self.num_states), fill_value=-1)
        self._output_edges = np.full((self.num_states, self.num_states), fill_value=-1)
        for state_from in range(self.num_states):
            for x, state_to in enumerate(self.next_states[state_from, :]):
                self._input_edges[state_from, state_to] = x
                self._output_edges[state_from, state_to] = self.outputs[state_from, x]

    def __repr__(self) -> str:
        args = ", ".join([
            f"next_states={self.next_states.tolist()}",
            f"outputs={self.outputs.tolist()}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def num_states(self) -> int:
        r"""
        The number of states of the machine.
        """
        return self.next_states.shape[0]

    @cached_property
    def num_input_symbols(self) -> int:
        r"""
        The size (cardinality) of the input alphabet $\mathcal{X}$.
        """
        return self.next_states.shape[1]

    @cached_property
    def num_output_symbols(self) -> int:
        r"""
        The size (cardinality) of the output alphabet $\mathcal{Y}$.
        """
        return int(np.amax(self.outputs))

    @cached_property
    def input_edges(self) -> npt.NDArray[np.integer]:
        r"""
        The matrix of input edges of the machine. It has shape $|\mathcal{S}| \times |\mathcal{S}|$. If there is an edge from $s_0 \in \mathcal{S}$ to $s_1 \in \mathcal{S}$, then the element in row $s_0$ and column $s_1$ is the input associated with that edge (an element of $\mathcal{X}$); if there is no such edge, then the element is $-1$.

        Examples:
            >>> fsm = komm.FiniteStateMachine(next_states=[[0,1], [2,3], [0,1], [2,3]], outputs=[[0,3], [1,2], [3,0], [2,1]])
            >>> fsm.input_edges
            array([[ 0,  1, -1, -1],
                   [-1, -1,  0,  1],
                   [ 0,  1, -1, -1],
                   [-1, -1,  0,  1]])
        """
        return self._input_edges

    @cached_property
    def output_edges(self) -> npt.NDArray[np.integer]:
        r"""
        The matrix of output edges of the machine. It has shape $|\mathcal{S}| \times |\mathcal{S}|$. If there is an edge from $s_0 \in \mathcal{S}$ to $s_1 \in \mathcal{S}$, then the element in row $s_0$ and column $s_1$ is the output associated with that edge (an element of $\mathcal{Y}$); if there is no such edge, then the element is $-1$.

        Examples:
            >>> fsm = komm.FiniteStateMachine(next_states=[[0,1], [2,3], [0,1], [2,3]], outputs=[[0,3], [1,2], [3,0], [2,1]])
            >>> fsm.output_edges
            array([[ 0,  3, -1, -1],
                   [-1, -1,  1,  2],
                   [ 3,  0, -1, -1],
                   [-1, -1,  2,  1]])
        """
        return self._output_edges

    def process(
        self,
        input_sequence: npt.ArrayLike,
        initial_state: int,
    ) -> tuple[npt.NDArray[np.integer], int]:
        r"""
        Returns the output sequence corresponding to a given input sequence. It assumes the machine starts at a given initial state $s_\mathrm{i}$. The input sequence and the output sequence are denoted by $\mathbf{x} = (x_0, x_1, \ldots, x_{L-1}) \in \mathcal{X}^L$ and $\mathbf{y} = (y_0, y_1, \ldots, y_{L-1}) \in \mathcal{Y}^L$, respectively.

        Parameters:
            input_sequence: The input sequence $\mathbf{x} \in \mathcal{X}^L$. It should be a 1D-array with elements in $\mathcal{X}$.

            initial_state: The initial state $s_\mathrm{i}$ of the machine. Should be an integer in $\mathcal{S}$.

        Returns:
            output_sequence: The output sequence $\mathbf{y} \in \mathcal{Y}^L$ corresponding to `input_sequence`, assuming the machine starts at the state given by `initial_state`. It is a 1D-array with elements in $\mathcal{Y}$.

            final_state: The final state $s_\mathrm{f}$ of the machine. It is an integer in $\mathcal{S}$.

        Examples:
            >>> fsm = komm.FiniteStateMachine(next_states=[[0,1], [2,3], [0,1], [2,3]], outputs=[[0,3], [1,2], [3,0], [2,1]])
            >>> input_sequence, initial_state = [1, 1, 0, 1, 0], 0
            >>> output_sequence, final_state = fsm.process(input_sequence, initial_state)
            >>> output_sequence
            array([3, 2, 2, 0, 1])
            >>> final_state
            2
        """
        output_sequence = np.empty_like(input_sequence, dtype=int)
        s = initial_state
        for t, x in np.ndenumerate(input_sequence):
            y = self.outputs[s, x]
            s = self.next_states[s, x]
            output_sequence[t] = y
        final_state = int(s)
        return output_sequence, final_state

    def viterbi(
        self,
        observed_sequence: Sequence[Z] | npt.NDArray[Any],
        metric_function: MetricFunction[Z],
        initial_metrics: npt.ArrayLike | None = None,
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.floating]]:
        r"""
        Applies the Viterbi algorithm on a given observed sequence. The Viterbi algorithm finds the most probable input sequence $\hat{\mathbf{x}}(s) \in \mathcal{X}^L$ ending in state $s$, for all $s \in \mathcal{S}$, given an observed sequence $\mathbf{z} \in \mathcal{Z}^L$. It is assumed uniform input priors. See <cite>LC04, Sec. 12.1</cite>.

        Parameters:
            observed_sequence: The observed sequence $\mathbf{z} \in \mathcal{Z}^L$.

            metric_function: The metric function $\mathcal{Y} \times \mathcal{Z} \to \mathbb{R}$.

            initial_metrics: The initial metrics for each state. It must be a 1D-array of length $|\mathcal{S}|$. The default value is `0.0` for all states.

        Returns:
            input_sequences_hat: The most probable input sequence $\hat{\mathbf{x}}(s) \in \mathcal{X}^L$ ending in state $s$, for all $s \in \mathcal{S}$. It is a 2D-array of shape $L \times |\mathcal{S}|$, in which column $s$ is equal to $\hat{\mathbf{x}}(s)$.

            final_metrics: The final metrics for each state. It is a 1D-array of length $|\mathcal{S}|$.
        """
        L, num_states = len(observed_sequence), self.num_states
        choices = np.empty((L, num_states), dtype=int)
        metrics = np.full((L + 1, num_states), fill_value=np.inf)
        if initial_metrics is None:
            metrics[0, :] = np.zeros(num_states, dtype=float)
        else:
            metrics[0, :] = initial_metrics
        for t, z in enumerate(observed_sequence):
            for s0 in range(num_states):
                for s1, y in zip(self.next_states[s0], self.outputs[s0]):
                    candidate_metrics = metrics[t, s0] + metric_function(y, z)
                    if candidate_metrics < metrics[t + 1, s1]:
                        metrics[t + 1, s1] = candidate_metrics
                        choices[t, s1] = s0

        # Backtrack
        input_sequences_hat = np.empty((L, num_states), dtype=int)
        for final_state in range(num_states):
            s1 = final_state
            for t in reversed(range(L)):
                s0 = choices[t, s1]
                input_sequences_hat[t, final_state] = self.input_edges[s0, s1]
                s1 = s0

        return input_sequences_hat, metrics[L, :]

    def viterbi_streaming(
        self,
        observed_sequence: Sequence[Z] | npt.NDArray[Any],
        metric_function: MetricFunction[Z],
        memory: MetricMemory,
    ) -> npt.NDArray[np.integer]:
        r"""
        Applies the streaming version of the Viterbi algorithm on a given observed sequence. The path memory (or traceback length) is denoted by $\tau$. It chooses the survivor with best metric and selects the information block on this path. See <cite>LC04, Sec. 12.3</cite>.

        Parameters:
            observed_sequence: The observed sequence $\mathbf{z} \in \mathcal{Z}^L$.

            metric_function: The metric function $\mathcal{Y} \times \mathcal{Z} \to \mathbb{R}$.

            memory: The metrics for each state. It must be a dictionary containing two keys: `'paths'`, a 2D-array of integers of shape $|\mathcal{S}| \times (\tau + 1)$; and `'metrics'`, a 2D-array of floats of shape $|\mathcal{S}| \times (\tau + 1)$. This dictionary is updated in-place by this method.

        Returns:
            input_sequence_hat: The most probable input sequence $\hat{\mathbf{x}} \in \mathcal{X}^L$
        """
        num_states = self.num_states
        input_sequences_hat = np.empty(len(observed_sequence), dtype=int)
        for t, z in enumerate(observed_sequence):
            new_metrics = np.full(num_states, fill_value=np.inf)
            choices = np.zeros(num_states, dtype=int)
            for s0 in range(num_states):
                for s1, y in zip(self.next_states[s0], self.outputs[s0]):
                    candidate_metric = memory["metrics"][s0, -1] + metric_function(y, z)
                    if candidate_metric < new_metrics[s1]:
                        new_metrics[s1] = candidate_metric
                        choices[s1] = s0

            s_star = np.argmin(new_metrics)
            s0, s1 = memory["paths"][s_star, :2]
            input_sequences_hat[t] = self.input_edges[s0, s1]

            memory["metrics"] = np.roll(memory["metrics"], shift=-1, axis=1)
            memory["metrics"][:, -1] = new_metrics
            memory["paths"] = np.roll(memory["paths"], shift=-1, axis=1)

            paths_copy = np.copy(memory["paths"])
            for s1, s0 in enumerate(choices):
                memory["paths"][s1, :-1] = paths_copy[s0, :-1]
                memory["paths"][s1, -1] = s1

        return input_sequences_hat

    def forward_backward(
        self,
        observed_sequence: Sequence[Z] | npt.NDArray[Any],
        metric_function: MetricFunction[Z],
        input_priors: npt.ArrayLike | None = None,
        initial_state_distribution: npt.ArrayLike | None = None,
        final_state_distribution: npt.ArrayLike | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Applies the forward-backward algorithm on a given observed sequence. The forward-backward algorithm computes the posterior pmf of each input $x_0, x_1, \ldots, x_{L-1} \in \mathcal{X}$ given an observed sequence $\mathbf{z} = (z_0, z_1, \ldots, z_{L-1}) \in \mathcal{Z}^L$. The prior pmf of each input may also be provided. See <cite>LC04, 12.6</cite>.

        Parameters:
            observed_sequence: The observed sequence $\mathbf{z} \in \mathcal{Z}^L$.

            metric_function: The metric function $\mathcal{Y} \times \mathcal{Z} \to \mathbb{R}$.

            input_priors: The prior pmf of each input, of shape $L \times |\mathcal{X}|$. The element in row $t \in [0 : L)$ and column $x \in \mathcal{X}$ should be $p(x_t = x)$. The default value yields uniform priors.

            initial_state_distribution: The pmf of the initial state of the machine. It must be a 1D-array of length $|\mathcal{S}|$. The default value is uniform over all states.

            final_state_distribution: The pmf of the final state of the machine. It must be a 1D-array of length $|\mathcal{S}|$. The default value is uniform over all states.

        Returns:
            input_posteriors: The posterior pmf of each input, given the observed sequence, of shape $L \times |\mathcal{X}|$. The element in row $t \in [0 : L)$ and column $x \in \mathcal{X}$ is $p(x_t = x \mid \mathbf{z})$.
        """
        L, num_states, num_input_symbols = (
            len(observed_sequence),
            self.num_states,
            self.num_input_symbols,
        )

        if input_priors is None:
            input_priors = np.ones((L, num_input_symbols)) / num_input_symbols
        if initial_state_distribution is None:
            initial_state_distribution = np.ones(num_states) / num_states
        if final_state_distribution is None:
            final_state_distribution = np.ones(num_states) / num_states

        log_gamma = np.full((L, num_states, num_states), fill_value=-np.inf)
        log_alpha = np.full((L + 1, num_states), fill_value=-np.inf)
        log_beta = np.full((L + 1, num_states), fill_value=-np.inf)

        with np.errstate(divide="ignore"):
            log_input_priors = np.log(input_priors)
            log_alpha[0, :] = np.log(initial_state_distribution)
            log_beta[L, :] = np.log(final_state_distribution)

        for t, z in enumerate(observed_sequence):
            for x, s0 in np.ndindex(num_input_symbols, num_states):
                y, s1 = self.outputs[s0, x], self.next_states[s0, x]
                log_gamma[t, s0, s1] = log_input_priors[t, x] + metric_function(y, z)

        for t in range(0, L - 1):
            for s1 in range(num_states):
                log_alpha[t + 1, s1] = np.logaddexp.reduce(
                    log_gamma[t, :, s1] + log_alpha[t, :]
                )

        for t in range(L - 1, -1, -1):
            for s0 in range(num_states):
                log_beta[t, s0] = np.logaddexp.reduce(
                    log_gamma[t, s0, :] + log_beta[t + 1, :]
                )

        log_input_posteriors = np.empty((L, num_input_symbols), dtype=float)
        edge_labels = np.empty(num_states, dtype=float)
        for t in range(L):
            for x in range(num_input_symbols):
                for s0 in range(num_states):
                    s1 = self.next_states[s0, x]
                    edge_labels[s0] = (
                        log_alpha[t, s0] + log_gamma[t, s0, s1] + log_beta[t + 1, s1]
                    )
                log_input_posteriors[t, x] = np.logaddexp.reduce(edge_labels)

        input_posteriors = np.exp(
            log_input_posteriors - np.amax(log_input_posteriors, axis=1, keepdims=True)
        )
        input_posteriors /= np.sum(input_posteriors, axis=1, keepdims=True)

        return input_posteriors



from functools import cache, cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from .._util.decorators import blockwise
from . import base


class SystematicBlockCode(base.BlockCode):
    r"""
    Systematic linear block code. A *systematic linear block code* is a [linear block code](/ref/BlockCode) in which the information bits can be found in predefined positions in the codeword, called the *information set* $\mathcal{K}$, which is a $k$-sublist of $[0 : n)$; the remaining positions are called the *parity set* $\mathcal{M}$, which is a $m$-sublist of $[0 : n)$. In this case, the generator matrix then has the property that the columns indexed by $\mathcal{K}$ are equal to $I_k$, and the columns indexed by $\mathcal{M}$ are equal to $P$. The check matrix has the property that the columns indexed by $\mathcal{M}$ are equal to $I_m$, and the columns indexed by $\mathcal{K}$ are equal to $P^\transpose$. The matrix $P \in \mathbb{B}^{k \times m}$ is called the *parity submatrix* of the code.

    The constructor expects the parity submatrix and the information set.

    Attributes:
        parity_submatrix: The parity submatrix $P$ the code, which is a $k \times m$ binary matrix.

        information_set: Either an array containing the indices of the information positions, which must be a $k$-sublist of $[0 : n)$, or one of the strings `'left'` or `'right'`. The default value is `'left'`.

    Examples:
        >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
        >>> (code.length, code.dimension, code.redundancy)
        (5, 2, 3)
        >>> code.generator_matrix
        array([[1, 0, 0, 1, 1],
               [0, 1, 1, 1, 0]])
        >>> code.check_matrix
        array([[0, 1, 1, 0, 0],
               [1, 1, 0, 1, 0],
               [1, 0, 0, 0, 1]])

        >>> code = komm.SystematicBlockCode(
        ...     parity_submatrix=[[0, 1, 1], [1, 1, 0]],
        ...     information_set='right',
        ... )
        >>> (code.length, code.dimension, code.redundancy)
        (5, 2, 3)
        >>> code.generator_matrix
        array([[0, 1, 1, 1, 0],
               [1, 1, 0, 0, 1]])
        >>> code.check_matrix
        array([[1, 0, 0, 0, 1],
               [0, 1, 0, 1, 1],
               [0, 0, 1, 1, 0]])
    """

    def __init__(
        self,
        parity_submatrix: npt.ArrayLike,
        information_set: Literal["left", "right"] | npt.ArrayLike = "left",
    ):
        self.parity_submatrix = np.asarray(parity_submatrix)
        n, k, m = self.length, self.dimension, self.redundancy
        if isinstance(information_set, str):
            if information_set == "left":
                self.information_set = np.arange(k)
            elif information_set == "right":
                self.information_set = np.arange(m, n)
            else:
                raise ValueError(
                    "if string, 'information_set' must be 'left' or 'right'"
                )
            return
        self.information_set = np.asarray(information_set)
        if (
            self.information_set.size != k
            or self.information_set.min() < 0
            or self.information_set.max() > n
        ):
            raise ValueError("'information_set' must be a 'k'-sublist of 'range(n)'")

    def __repr__(self) -> str:
        args = ", ".join([
            f"parity_submatrix={self.parity_submatrix.tolist()}",
            f"information_set={self.information_set.tolist()}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def parity_set(self) -> npt.NDArray[np.integer]:
        return np.setdiff1d(np.arange(self.length), self.information_set)

    @cached_property
    def length(self) -> int:
        r"""
        Examples:
           >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
           >>> code.length
           5
        """
        return self.dimension + self.redundancy

    @cached_property
    def dimension(self) -> int:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.dimension
            2
        """
        return self.parity_submatrix.shape[0]

    @cached_property
    def redundancy(self) -> int:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.redundancy
            3
        """
        return self.parity_submatrix.shape[1]

    @cached_property
    def rate(self) -> float:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.rate
            0.4
        """
        return super().rate

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.generator_matrix
            array([[1, 0, 0, 1, 1],
                   [0, 1, 1, 1, 0]])
        """
        k, n = self.dimension, self.length
        matrix = np.empty((k, n), dtype=int)
        matrix[:, self.information_set] = np.eye(k, dtype=int)
        matrix[:, self.parity_set] = self.parity_submatrix
        return matrix

    @cached_property
    def generator_matrix_right_inverse(self) -> npt.NDArray[np.integer]:
        k, n = self.dimension, self.length
        matrix = np.zeros((n, k), dtype=int)
        matrix[self.information_set] = np.eye(k, dtype=int)
        return matrix

    @cached_property
    def check_matrix(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.check_matrix
            array([[0, 1, 1, 0, 0],
                   [1, 1, 0, 1, 0],
                   [1, 0, 0, 0, 1]])
        """
        m, n = self.redundancy, self.length
        matrix = np.empty((m, n), dtype=int)
        matrix[:, self.information_set] = self.parity_submatrix.T
        matrix[:, self.parity_set] = np.eye(m, dtype=int)
        return matrix

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.generator_matrix
            array([[1, 0, 0, 1, 1],
                   [0, 1, 1, 1, 0]])
            >>> code.encode([0, 0])  # Sequence with single message
            array([0, 0, 0, 0, 0])
            >>> code.encode([0, 0, 1, 1])  # Sequence with two messages
            array([0, 0, 0, 0, 0, 1, 1, 1, 0, 1])
            >>> code.encode([[0, 0],  # 2D array of single messages
            ...              [1, 1]])
            array([[0, 0, 0, 0, 0],
                   [1, 1, 1, 0, 1]])
            >>> code.encode([[0, 0, 1, 1],  # 2D array of two messages
            ...              [1, 1, 1, 0]])
            array([[0, 0, 0, 0, 0, 1, 1, 1, 0, 1],
                   [1, 1, 1, 0, 1, 1, 0, 0, 1, 1]])
        """

        @blockwise(self.dimension)
        def encode(u: npt.NDArray[np.integer]) -> npt.NDArray[np.integer]:
            v = np.empty(u.shape[:-1] + (self.length,), dtype=int)
            v[..., self.information_set] = u
            v[..., self.parity_set] = u @ self.parity_submatrix % 2
            return v

        return encode(input)

    def project_word(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        @blockwise(self.length)
        def project(v: npt.NDArray[np.integer]) -> npt.NDArray[np.integer]:
            u = v[..., self.information_set]
            return u

        return project(input)

    def inverse_encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.inverse_encode([0, 0, 0, 0, 0])  # Sequence with single codeword
            array([0, 0])
            >>> code.inverse_encode([0, 0, 0, 0, 0, 1, 1, 1, 0, 1])  # Sequence with two codewords
            array([0, 0, 1, 1])
            >>> code.inverse_encode([[0, 0, 0, 0, 0],  # 2D array of single codewords
            ...                      [1, 1, 1, 0, 1]])
            array([[0, 0],
                   [1, 1]])
            >>> code.inverse_encode([[0, 0, 0, 0, 0, 1, 1, 1, 0, 1],  # 2D array of two codewords
            ...                      [1, 1, 1, 0, 1, 1, 0, 0, 1, 1]])
            array([[0, 0, 1, 1],
                   [1, 1, 1, 0]])
        """
        return super().inverse_encode(input)

    def check(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.check([1, 1, 1, 0, 1])  # Sequence with single received word
            array([0, 0, 0])
            >>> code.check([1, 1, 1, 0, 1, 1, 1, 1, 1, 1])  # Sequence with two received words
            array([0, 0, 0, 0, 1, 0])
            >>> code.check([[1, 1, 1, 0, 1],  # 2D array of single received words
            ...             [1, 1, 1, 1, 1]])
            array([[0, 0, 0],
                   [0, 1, 0]])
            >>> code.check([[1, 1, 1, 0, 1, 1, 1, 1, 1, 1],  # 2D array of two received words
            ...             [1, 1, 1, 1, 1, 0, 0, 0, 1, 1]])
            array([[0, 0, 0, 0, 1, 0],
                   [0, 1, 0, 0, 1, 1]])
        """

        @blockwise(self.length)
        def check(r: npt.NDArray[np.integer]) -> npt.NDArray[np.integer]:
            r_inf = r[..., self.information_set]
            r_par = r[..., self.parity_set]
            s = (r_inf @ self.parity_submatrix + r_par) % 2
            return s

        return check(input)

    @cache
    def codewords(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.codewords()
            array([[0, 0, 0, 0, 0],
                   [1, 0, 0, 1, 1],
                   [0, 1, 1, 1, 0],
                   [1, 1, 1, 0, 1]])
        """
        return super().codewords()

    @cache
    def codeword_weight_distribution(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.codeword_weight_distribution()
            array([1, 0, 0, 2, 1, 0])
        """
        return super().codeword_weight_distribution()

    @cache
    def minimum_distance(self) -> int:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.minimum_distance()
            3
        """
        return super().minimum_distance()

    @cache
    def coset_leaders(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.coset_leaders()
            array([[0, 0, 0, 0, 0],
                   [0, 0, 1, 0, 0],
                   [0, 0, 0, 1, 0],
                   [0, 1, 0, 0, 0],
                   [0, 0, 0, 0, 1],
                   [1, 1, 0, 0, 0],
                   [1, 0, 0, 0, 0],
                   [1, 0, 1, 0, 0]])
        """
        return super().coset_leaders()

    @cache
    def coset_leader_weight_distribution(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.coset_leader_weight_distribution()
            array([1, 5, 2, 0, 0, 0])
        """
        return super().coset_leader_weight_distribution()

    @cache
    def packing_radius(self) -> int:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.packing_radius()
            1
        """
        return super().packing_radius()

    @cache
    def covering_radius(self) -> int:
        r"""
        Examples:
            >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 1, 0]])
            >>> code.covering_radius()
            2
        """
        return super().covering_radius()



from functools import cache, cached_property
from itertools import count, product

import numpy as np
import numpy.typing as npt
from typing_extensions import Self

from .._util.information_theory import PMF
from .util import (
    Word,
    is_fully_covering,
    is_prefix_free,
    is_uniquely_parsable,
    parse_fixed_length,
    parse_prefix_free,
)


class VariableToFixedCode:
    r"""
    General variable-to-fixed length code. A *variable-to-fixed length code* with *target alphabet* $\mathcal{T}$, *source alphabet* $\mathcal{S}$, and *target block size* $n$ is defined by a (possibly partial) decoding mapping $\mathrm{Dec} : \mathcal{T}^n \rightharpoonup \mathcal{S}^+$, where the domain is the set of all $n$-tuples with entries in $\mathcal{T}$, and the co-domain is the set of all finite-length, non-empty tuples with entries in $\mathcal{S}$. Here, we assume that $\mathcal{T} = [0:T)$ and $\mathcal{S} = [0:S)$, for integers $T \geq 2$ and $S \geq 2$. The elements in the image of $\mathrm{Dec}$ are called *sourcewords*.
    """

    def __init__(
        self,
        target_cardinality: int,
        source_cardinality: int,
        target_block_size: int,
        dec_mapping: dict[Word, Word],
    ) -> None:
        self._target_cardinality = target_cardinality
        self._source_cardinality = source_cardinality
        self._target_block_size = target_block_size
        self._dec_mapping = dec_mapping
        self.__post_init__()

    def __post_init__(self) -> None:
        domain, codomain = self.dec_mapping.keys(), self.dec_mapping.values()
        T = self.target_cardinality
        S = self.source_cardinality
        n = self.target_block_size
        if not T >= 2:
            raise ValueError("'target_cardinality': must be at least 2")
        if not S >= 2:
            raise ValueError("'source_cardinality': must be at least 2")
        if not n >= 1:
            raise ValueError("'target_block_size': must be at least 1")
        if not set(domain) <= set(product(range(T), repeat=n)):
            raise ValueError(f"'dec_mapping': invalid domain")
        if not all(
            all(0 <= x < S for x in word) and len(word) > 0 for word in codomain
        ):
            raise ValueError(f"'dec_mapping': invalid co-domain")

    def __repr__(self) -> str:
        args = ", ".join([
            f"target_cardinality={self.target_cardinality}",
            f"source_cardinality={self.source_cardinality}",
            f"target_block_size={self.target_block_size}",
            f"dec_mapping={self.dec_mapping}",
        ])
        return f"{__class__.__name__}({args})"

    @classmethod
    def from_dec_mapping(cls, dec_mapping: dict[Word, Word]) -> Self:
        r"""
        Constructs a variable-to-fixed length code from the decoding mapping $\Dec$.

        Parameters:
            dec_mapping: The decoding mapping $\Dec$. Must be a dictionary of length at most $S^n$ whose keys are $n$-tuples of integers in $[0:T)$ and whose values are non-empty tuples of integers in $[0:S)$.

        Notes:
            The target block size $n$ is inferred from the domain of the decoding mapping, and the target and source cardinalities $T$ and $S$ are inferred from the maximum values in the domain and co-domain, respectively.

        Examples:
            >>> code = komm.VariableToFixedCode.from_dec_mapping({
            ...     (0, 0): (0, 0, 0),
            ...     (0, 1): (0, 0, 1),
            ...     (1, 0): (0, 1),
            ...     (1, 1): (1,),
            ... })
            >>> code.target_cardinality, code.source_cardinality, code.target_block_size
            (2, 2, 2)
            >>> code.sourcewords
            [(0, 0, 0), (0, 0, 1), (0, 1), (1,)]

            >>> code = komm.VariableToFixedCode.from_dec_mapping({
            ...     (0, 0, 0): (1, ),
            ...     (0, 0, 1): (2, ),
            ...     (0, 1, 0): (0, 1),
            ...     (0, 1, 1): (0, 2),
            ...     (1, 0, 0): (0, 0, 0),
            ...     (1, 0, 1): (0, 0, 1),
            ...     (1, 1, 0): (0, 0, 2),
            ... })  # Incomplete mapping
            >>> code.target_cardinality, code.source_cardinality, code.target_block_size
            (2, 3, 3)
            >>> code.sourcewords
            [(1,), (2,), (0, 1), (0, 2), (0, 0, 0), (0, 0, 1), (0, 0, 2)]
        """
        domain, codomain = dec_mapping.keys(), dec_mapping.values()
        T = max(max(word) for word in domain) + 1
        S = max(max(word) for word in codomain) + 1
        n = len(next(iter(domain)))
        return cls(T, S, n, dec_mapping)

    @classmethod
    def from_sourcewords(cls, target_cardinality: int, sourcewords: list[Word]) -> Self:
        r"""
        Constructs a variable-to-fixed length code from the target cardinality $T$ and a list of sourcewords.

        Parameters:
            target_cardinality: The target cardinality $T$. Must be an integer greater than or equal to $2$.

            sourcewords: The sourcewords of the code. Must be a list of length at most $T^n$ containing tuples of integers in $[0:S)$, where $S$ is the source cardinality of the code. The tuple in position $i$ must be equal to $\mathrm{Dec}(v)$, where $v$ is the $i$-th element in the lexicographic ordering of $[0:T)^n$.

        Note:
            The target block size $n$ is inferred from the length of the sourcewords, and the source cardinality $S$ is inferred from the maximum value in the sourcewords.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(
            ...     target_cardinality=2,
            ...     sourcewords=[(0,0,0), (0,0,1), (0,1), (1,)],
            ... )
            >>> code.target_cardinality, code.source_cardinality, code.target_block_size
            (2, 2, 2)
            >>> code.dec_mapping
            {(0, 0): (0, 0, 0),
             (0, 1): (0, 0, 1),
             (1, 0): (0, 1),
             (1, 1): (1,)}

            >>> code = komm.VariableToFixedCode.from_sourcewords(
            ...     target_cardinality=2,
            ...     sourcewords=[(1,), (2,), (0,1), (0,2), (0,0,0), (0,0,1), (0,0,2)],
            ... )
            >>> code.target_cardinality, code.source_cardinality, code.target_block_size
            (2, 3, 3)
            >>> code.dec_mapping
            {(0, 0, 0): (1,),
             (0, 0, 1): (2,),
             (0, 1, 0): (0, 1),
             (0, 1, 1): (0, 2),
             (1, 0, 0): (0, 0, 0),
             (1, 0, 1): (0, 0, 1),
             (1, 1, 0): (0, 0, 2)}
        """
        T = target_cardinality
        S = max(max(word) for word in sourcewords) + 1
        n = next(n for n in count(1) if T**n >= len(sourcewords))
        dec_mapping = dict(zip(product(range(T), repeat=n), sourcewords))
        return cls(T, S, n, dec_mapping)

    @cached_property
    def target_cardinality(self) -> int:
        r"""
        The target cardinality $T$ of the code. It is the number of symbols in the target alphabet.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.target_cardinality
            2
        """
        return self._target_cardinality

    @cached_property
    def source_cardinality(self) -> int:
        r"""
        The source cardinality $S$ of the code. It is the number of symbols in the source alphabet.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.source_cardinality
            2
        """
        return self._source_cardinality

    @cached_property
    def target_block_size(self) -> int:
        r"""
        The target block size $n$ of the code. It is the number of symbols in each target block.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.target_block_size
            2
        """
        return self._target_block_size

    @cached_property
    def size(self) -> int:
        r"""
        The number of sourcewords in the code. It is less than or equal to $T^n$.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.size
            3
        """
        return len(self.dec_mapping)

    @cached_property
    def dec_mapping(self) -> dict[Word, Word]:
        r"""
        The decoding mapping $\mathrm{Dec}$ of the code. It is a dictionary of length at most $T^n$ whose keys are $n$-tuples of integers in $[0:T)$ and whose values are the corresponding sourcewords.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.dec_mapping
            {(0, 0): (0,), (0, 1): (1,), (1, 0): (0, 1)}
        """
        return self._dec_mapping

    @cached_property
    def sourcewords(self) -> list[Word]:
        r"""
        The sourcewords of the code. They correspond to the image of the decoding mapping $\mathrm{Dec}$.

        Examples:
            >>> code = komm.VariableToFixedCode.from_dec_mapping({
            ...     (0, 0): (0,),
            ...     (0, 1): (1,),
            ...     (1, 0): (0, 1),
            ... })
            >>> code.sourcewords
            [(0,), (1,), (0, 1)]
        """
        return list(self.dec_mapping.values())

    @cached_property
    def _inv_dec_mapping(self) -> dict[Word, Word]:
        return {v: k for k, v in self.dec_mapping.items()}

    @cache
    def is_fully_covering(self) -> bool:
        """
        Returns whether the code is fully covering. A code is *fully covering* if every possible source sequence has a prefix that is a sourceword.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,0), (1,1)])
            >>> code.is_fully_covering()
            True

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.is_fully_covering()
            True

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0), (0,1)])
            >>> code.is_fully_covering()  # (1,) is not covered
            False

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (0,1)])
            >>> code.is_fully_covering()  # (1,) is not covered
            False
        """
        return is_fully_covering(self.sourcewords, self.source_cardinality)

    @cache
    def is_uniquely_encodable(self) -> bool:
        r"""
        Returns whether the code is uniquely encodable. A code is *uniquely encodable* if there is a unique way to parse any concatenation of sourcewords.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,0), (1,1)])
            >>> code.is_uniquely_encodable()
            True

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.is_uniquely_encodable()  # 01 can be parsed as 0|1 or 01
            False

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0), (0,1)])
            >>> code.is_uniquely_encodable()
            True

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (0,1)])
            >>> code.is_uniquely_encodable()
            True
        """
        return is_uniquely_parsable(self.sourcewords)

    @cache
    def is_prefix_free(self) -> bool:
        r"""
        Returns whether the code is prefix-free. A code is *prefix-free* if no sourceword is a prefix of any other sourceword.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,0), (1,1)])
            >>> code.is_prefix_free()
            True

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.is_prefix_free()
            False

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0), (0,1)])
            >>> code.is_prefix_free()
            True

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (0,1)])
            >>> code.is_prefix_free()
            False
        """
        return is_prefix_free(self.sourcewords)

    def rate(self, pmf: npt.ArrayLike) -> float:
        r"""
        Computes the expected rate $R$ of the code, considering a given pmf. This quantity is given by
        $$
            R = \frac{n}{\bar{k}},
        $$
        where $n$ is the target block size, and $\bar{k}$ is the expected sourceword length, assuming iid source symbols drawn from $p_X$. It is measured in $T$-ary digits per source symbol.

        Parameters:
            pmf: The (first-order) probability mass function $p_X$ to be considered.

        Returns:
            rate: The expected rate $R$ of the code.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.rate([2/3, 1/3])
            np.float64(1.3846153846153846)
        """
        pmf = PMF(pmf)
        n = self.target_block_size
        probabilities = [np.prod([pmf[x] for x in word]) for word in self.sourcewords]
        lengths = [len(word) for word in self.sourcewords]
        return n / np.dot(lengths, probabilities)

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Encodes a sequence of source symbols using the code, which must be fully covering and uniquely encodable. When the input sequence ends with symbols that form only a partial match with any sourceword, the encoder will complete this last block using any valid sourceword that starts with these remaining symbols.

        Warning:
            Encoding for non-prefix-free codes is not implemented yet.

        Parameters:
            input: The sequence of symbols to be encoded. Must be a 1D-array with elements in $[0:S)$ (where $S$ is the source cardinality of the code).

        Returns:
            output: The sequence of encoded symbols. It is a 1D-array with elements in $[0:T)$ (where $T$ is the target cardinality of the code) with a length that is a multiple of the target block size $n$.

        Examples:
            >>> code = komm.VariableToFixedCode.from_dec_mapping({
            ...     (0, 0): (0, 0, 0),
            ...     (0, 1): (0, 0, 1),
            ...     (1, 0): (0, 1),
            ...     (1, 1): (1,),
            ... })

            >>> code.encode([1, 0, 0, 0])  # Parsed as 1|000
            array([1, 1, 0, 0])

            >>> code.encode([1, 0, 0])  # Incomplete input, completed as 1|000
            array([1, 1, 0, 0])

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0), (0,1)])
            >>> code.encode([1, 0, 0, 0])  # Code is not fully covering
            Traceback (most recent call last):
            ...
            ValueError: code is not fully covering

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.encode([1, 0, 0, 0])  # Code is not uniquely encodable
            Traceback (most recent call last):
            ...
            ValueError: code is not uniquely encodable
        """
        if not self.is_fully_covering():
            raise ValueError("code is not fully covering")
        if not self.is_uniquely_encodable():
            raise ValueError("code is not uniquely encodable")
        if not self.is_prefix_free():
            raise NotImplementedError(
                "encoding for non-prefix-free codes is not implemented yet"
            )
        input = np.asarray(input)
        return parse_prefix_free(input, self._inv_dec_mapping, allow_incomplete=True)

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Decodes a sequence of target symbols using the code, which must be fully covering and uniquely encodable.

        Parameters:
            input: The sequence of symbols to be decoded. Must be a 1D-array with elements in $[0:T)$ (where $T$ is the target cardinality of the code) and have a length that is a multiple of the target block size $n$. Also, the sequence must be a concatenation of target words (i.e., the output of the `encode` method).

        Returns:
            The sequence of decoded symbols. It is a 1D-array with elements in $[0:S)$ (where $S$ is the source cardinality of the code).

        Examples:
            >>> code = komm.VariableToFixedCode.from_dec_mapping({
            ...     (0, 0): (0,),
            ...     (0, 1): (1,),
            ...     (1, 0): (2,),
            ... })
            >>> code.decode([0, 0, 1, 0])
            array([0, 2])

            >>> code.decode([1, 1, 0, 0, 1])  # Not a multiple of target block size
            Traceback (most recent call last):
            ...
            ValueError: length of input must be a multiple of block size 2 (got 5)

            >>> code.decode([0, 0, 1, 1])  # 11 is not a valid target word
            Traceback (most recent call last):
            ...
            ValueError: input contains invalid word

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0), (0,1)])
            >>> code.decode([0, 0, 0, 1])  # Code is not fully covering
            Traceback (most recent call last):
            ...
            ValueError: code is not fully covering

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,), (1,), (0,1)])
            >>> code.decode([0, 0, 0, 1])  # Code is not uniquely encodable
            Traceback (most recent call last):
            ...
            ValueError: code is not uniquely encodable
        """
        if not self.is_fully_covering():
            raise ValueError("code is not fully covering")
        if not self.is_uniquely_encodable():
            raise ValueError("code is not uniquely encodable")
        input = np.asarray(input)
        return parse_fixed_length(input, self.dec_mapping, self.target_block_size)



import operator
from dataclasses import dataclass
from functools import cached_property, reduce

from .._algebra.BinaryPolynomial import BinaryPolynomial
from .._algebra.FiniteBifield import FiniteBifield, FiniteBifieldElement
from .CyclicCode import CyclicCode


@dataclass(eq=False)
class BCHCode(CyclicCode):
    r"""
    Bose–Ray-Chaudhuri–Hocquenghem (BCH) code. For given parameters $\mu \geq 2$ and $\delta$ satisfying $2 \leq \delta \leq 2^{\mu} - 1$, a *binary BCH code* is a [cyclic code](/ref/CyclicCode) with generator polynomial given by
    $$
        g(X) = \mathrm{lcm} \left\\{ \phi_1(X), \phi_2(X), \ldots, \phi_{\delta - 1}(X) \right\\},
    $$
    where $\phi_i(X)$ is the minimal polynomial of $\alpha^i$, and $\alpha$ is a primitive element of $\mathrm{GF}(2^\mu)$. The parameter $\delta$ must be a *Bose distance*. The resulting code is denoted by $\bch(\mu, \delta)$, and has the following parameters, where $\delta = 2 \tau + 1$:

    - Length: $n = 2^{\mu} - 1$
    - Dimension: $k \geq n - \mu \tau$
    - Redundancy: $m \leq \mu \tau$
    - Minimum distance: $d \geq \delta$

    Only *narrow-sense* and *primitive* BCH codes are implemented. For more details, see <cite>LC04, Ch. 6</cite> and <cite>HP03, Sec. 5.1</cite>.

    The table below lists the possible values of $\delta$ for $2 \leq \mu \leq 10$.

    | $\mu$ | $n$ |  Bose distances $\delta$ |
    | :-: | :-: | --- |
    | $2$ | $3$ | $3$ |
    | $3$ | $7$ | $3$, $7$ |
    | $4$ | $15$ | $3$, $5$, $7$, $15$ |
    | $5$ | $31$ | $3$, $5$, $7$, $11$, $15$, $31$ |
    | $6$ | $63$ | $3$, $5$, $7$, $9$, $11$, $13$, $15$, $21$, $23$, $27$, $31$, $63$ |
    | $7$ | $127$ | $3$, $5$, $7$, $9$, $11$, $13$, $15$, $19$, $21$, $23$, $27$, $29$, $31$, $43$, $47$, $55$, $63$, $127$ |
    | $8$ | $255$ | $3$, $5$, $7$, $9$, $11$, $13$, $15$, $17$, $19$, $21$, $23$, $25$, $27$, $29$, $31$, $37$, $39$, $43$, $45$, $47$, $51$, $53$, $55$, $59$, $61$, $63$, $85$, $87$, $91$, $95$, $111$, $119$, $127$, $255$ |
    | $9$ | $511$ | $3$, $5$, $7$, $9$, $11$, $13$, $15$, $17$, $19$, $21$, $23$, $25$, $27$, $29$, $31$, $35$, $37$, $39$, $41$, $43$, $45$, $47$, $51$, $53$, $55$, $57$, $59$, $61$, $63$, $73$, $75$, $77$, $79$, $83$, $85$, $87$, $91$, $93$, $95$, $103$, $107$, $109$, $111$, $117$, $119$, $123$, $125$, $127$, $171$, $175$, $183$, $187$, $191$, $219$, $223$, $239$, $255$, $511$ |
    | $10$ | $1023$ | $3$, $5$, $7$, $9$, $11$, $13$, $15$, $17$, $19$, $21$, $23$, $25$, $27$, $29$, $31$, $33$, $35$, $37$, $39$, $41$, $43$, $45$, $47$, $49$, $51$, $53$, $55$, $57$, $59$, $61$, $63$, $69$, $71$, $73$, $75$, $77$, $79$, $83$, $85$, $87$, $89$, $91$, $93$, $95$, $99$, $101$, $103$, $105$, $107$, $109$, $111$, $115$, $117$, $119$, $121$, $123$, $125$, $127$, $147$, $149$, $151$, $155$, $157$, $159$, $165$, $167$, $171$, $173$, $175$, $179$, $181$, $183$, $187$, $189$, $191$, $205$, $207$, $213$, $215$, $219$, $221$, $223$, $231$, $235$, $237$, $239$, $245$, $247$, $251$, $253$, $255$, $341$, $343$, $347$, $351$, $363$, $367$, $375$, $379$, $383$, $439$, $447$, $479$, $495$, $511$, $1023$ |

    Notes:
        - For $\delta = 3$ it reduces to the [Hamming code](/ref/HammingCode).
        - For $\delta = 2^{\mu} - 1$ it reduces to the [repetition code](/ref/RepetitionCode).

    Attributes:
        mu: The parameter $\mu$ of the BCH code.
        delta: The Bose distance $\delta$ of the BCH code.

    Examples:
        >>> code = komm.BCHCode(mu=5, delta=7)
        >>> (code.length, code.dimension, code.redundancy)
        (31, 16, 15)
        >>> code.generator_polynomial
        BinaryPolynomial(0b1000111110101111)
        >>> code.minimum_distance()
        7

        >>> komm.BCHCode(mu=7, delta=31)
        BCHCode(mu=7, delta=31)

        >>> komm.BCHCode(mu=7, delta=32)
        Traceback (most recent call last):
        ...
        ValueError: 'delta' must be a Bose distance (the next one is 43)

        >>> komm.BCHCode(mu=7, delta=43)
        BCHCode(mu=7, delta=43)
    """

    mu: int
    delta: int

    def __post_init__(self) -> None:
        if not self.mu >= 2:
            raise ValueError("'mu' must satisfy mu >= 2")
        if not 2 <= self.delta <= 2**self.mu - 1:
            raise ValueError("'delta' must satisfy 2 <= delta <= 2**mu - 1")

        def phi(i: int):
            return (self.alpha**i).minimal_polynomial()

        lcm_set = {phi(i) for i in range(1, self.delta)}
        if phi(self.delta) in lcm_set:
            bose_distance = self.delta
            while phi(bose_distance) in lcm_set:
                bose_distance += 1
            raise ValueError(
                f"'delta' must be a Bose distance (the next one is {bose_distance})"
            )
        length = 2**self.mu - 1
        generator_polynomial = reduce(operator.mul, lcm_set)
        super().__init__(length=length, generator_polynomial=generator_polynomial)

    @cached_property
    def field(self) -> FiniteBifield:
        return FiniteBifield(self.mu)

    @cached_property
    def alpha(self) -> FiniteBifieldElement[FiniteBifield]:
        # Since the default modulus is a primitive polynomial, alpha = X is a primitive element.
        return self.field(0b10)

    def bch_syndrome(
        self, r_poly: BinaryPolynomial
    ) -> list[FiniteBifieldElement[FiniteBifield]]:
        # BCH syndrome computation. See [LC04, p. 205–209].
        return [r_poly.evaluate(self.alpha**i) for i in range(1, self.delta)]



from typing import SupportsInt

from typing_extensions import Self

from . import field
from .BinaryPolynomial import BinaryPolynomial


class BinaryPolynomialFraction:
    r"""
    Binary polynomial fraction. A *binary polynomial fraction* is a ratio of two [binary polynomials](/ref/BinaryPolynomial).

    Parameters:
        numerator: The numerator of the fraction.
        denominator: The denominator of the fraction.

    Examples:
        >>> komm.BinaryPolynomialFraction(0b11010, 0b101)  # (X^4 + X^3 + X) / (X^2 + 1)
        BinaryPolynomialFraction(0b11010, 0b101)

    <h2>Algebraic structure</h2>

    The binary polynomial fractions form a *field*. The following operations are supported: addition (`+`), subtraction (`-`), multiplication (`*`), division (`/`), and exponentiation (`**`).
    """

    numerator: BinaryPolynomial
    denominator: BinaryPolynomial

    def __init__(self, numerator: SupportsInt, denominator: SupportsInt = 0b1) -> None:
        self.numerator = BinaryPolynomial(numerator)
        self.denominator = BinaryPolynomial(denominator)
        self.__post_init__()

    def __post_init__(self):
        if self.denominator == 0b0:
            raise ZeroDivisionError("denominator cannot be zero")
        gcd = BinaryPolynomial.gcd(self.numerator, self.denominator)
        self.numerator //= gcd
        self.denominator //= gcd

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.numerator}, {self.denominator})"

    def __str__(self) -> str:
        return str(self.numerator) + "/" + str(self.denominator)

    @property
    def ambient(self):
        return BinaryPolynomialFractions()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.numerator == other.numerator and self.denominator == other.denominator
        )

    def __add__(self, other: Self) -> Self:
        n = self.numerator * other.denominator + other.numerator * self.denominator
        d = self.denominator * other.denominator
        return self.__class__(n, d)

    def __sub__(self, other: Self) -> Self:
        n = self.numerator * other.denominator - other.numerator * self.denominator
        d = self.denominator * other.denominator
        return self.__class__(n, d)

    def __neg__(self) -> Self:
        return self

    def __mul__(self, other: Self) -> Self:
        n = self.numerator * other.numerator
        d = self.denominator * other.denominator
        return self.__class__(n, d)

    def __rmul__(self, other: int) -> Self:
        return self.__class__(other * self.numerator, self.denominator)

    def __truediv__(self, other: Self) -> Self:
        n = self.numerator * other.denominator
        d = self.denominator * other.numerator
        return self.__class__(n, d)

    def __pow__(self, exponent: int) -> Self:
        return field.power(self, exponent)

    def inverse(self) -> Self:
        return self.__class__(self.denominator, self.numerator)


class BinaryPolynomialFractions:
    def __call__(self, value: tuple[int, int]) -> BinaryPolynomialFraction:
        return BinaryPolynomialFraction(*value)

    @property
    def zero(self) -> BinaryPolynomialFraction:
        return BinaryPolynomialFraction(0, 1)

    @property
    def one(self) -> BinaryPolynomialFraction:
        return BinaryPolynomialFraction(1, 1)

    @property
    def characteristic(self) -> int:
        return 2



from functools import cache

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from .._util.docs import mkdocstrings
from .._util.information_theory import PMF
from .FixedToVariableCode import FixedToVariableCode
from .util import Word, empty_mapping, extended_probabilities


@mkdocstrings(filters=["!.*"])
class FanoCode(FixedToVariableCode):
    r"""
    Binary Fano code. It is a [fixed-to-variable length code](/ref/FixedToVariableCode) in which the source words are first sorted in descending order of probability and then are recursively partitioned into two groups of approximately equal total probability, assigning bit $\mathtt{0}$ to one group and bit $\mathtt{1}$ to the other, until each source word is assigned a unique codeword. For more details, see [Wikipedia: Shannon–Fano coding](https://en.wikipedia.org/wiki/Shannon%E2%80%93Fano_coding).

    Notes:
        Fano codes are always [prefix-free](/ref/FixedToVariableCode/#is_prefix_free) (hence [uniquely decodable](/ref/FixedToVariableCode/#is_uniquely_decodable)).

    Parameters:
        pmf: The probability mass function of the source.
        source_block_size: The source block size $k$. The default value is $k = 1$.

    Examples:
        >>> pmf = [0.7, 0.15, 0.15]

        >>> code = komm.FanoCode(pmf, 1)
        >>> code.enc_mapping
        {(0,): (0,),
         (1,): (1, 0),
         (2,): (1, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.3)

        >>> code = komm.FanoCode(pmf, 2)
        >>> code.enc_mapping
        {(0, 0): (0,),
         (0, 1): (1, 0, 0),
         (0, 2): (1, 0, 1),
         (1, 0): (1, 1, 0),
         (1, 1): (1, 1, 1, 1, 0, 0),
         (1, 2): (1, 1, 1, 1, 0, 1),
         (2, 0): (1, 1, 1, 0),
         (2, 1): (1, 1, 1, 1, 1, 0),
         (2, 2): (1, 1, 1, 1, 1, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.1975)
    """

    def __init__(self, pmf: npt.ArrayLike, source_block_size: int = 1):
        self.pmf = PMF(pmf)
        super().__init__(
            source_cardinality=self.pmf.size,
            target_cardinality=2,
            source_block_size=source_block_size,
            enc_mapping=fano_algorithm(self.pmf, source_block_size),
        )

    def __repr__(self) -> str:
        args = ", ".join([
            f"pmf={self.pmf.tolist()}",
            f"source_block_size={self.source_block_size}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cache
    def is_uniquely_decodable(self) -> bool:
        return True

    @cache
    def is_prefix_free(self) -> bool:
        return True


def fano_algorithm(pmf: PMF, source_block_size: int) -> dict[Word, Word]:
    pbar = tqdm(
        desc="Generating Fano code",
        total=2 * pmf.size**source_block_size,
        delay=2.5,
    )

    enc_mapping = empty_mapping(pmf.size, source_block_size)
    group = extended_probabilities(pmf, source_block_size, pbar)
    stack: list[tuple[list[tuple[Word, float]], Word]] = [(group, ())]
    while stack:
        group, v = stack.pop()
        if len(group) == 1:
            u, _ = group[0]
            enc_mapping[u] = v
            pbar.update()
            continue
        probs = [p for _, p in group]
        total = np.sum(probs)
        index = np.argmin(np.abs(np.cumsum(probs) - total / 2))
        stack.append((group[index + 1 :], v + (1,)))
        stack.append((group[: index + 1], v + (0,)))

    pbar.close()

    return enc_mapping



from .BCHCode import BCHCode
from .BlockCode import BlockCode
from .CordaroWagnerCode import CordaroWagnerCode
from .CyclicCode import CyclicCode
from .GolayCode import GolayCode
from .HammingCode import HammingCode
from .Lexicode import Lexicode
from .ReedMullerCode import ReedMullerCode
from .RepetitionCode import RepetitionCode
from .SimplexCode import SimplexCode
from .SingleParityCheckCode import SingleParityCheckCode
from .SlepianArray import SlepianArray
from .SystematicBlockCode import SystematicBlockCode

__all__ = [
    "BCHCode",
    "BlockCode",
    "CordaroWagnerCode",
    "CyclicCode",
    "GolayCode",
    "HammingCode",
    "Lexicode",
    "ReedMullerCode",
    "RepetitionCode",
    "SimplexCode",
    "SingleParityCheckCode",
    "SlepianArray",
    "SystematicBlockCode",
]



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

import numpy as np
import numpy.typing as npt

from .._error_control_block.base import BlockCode

T = TypeVar("T", bound=BlockCode)


class BlockDecoder(ABC, Generic[T]):
    code: T

    @abstractmethod
    def __call__(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Decode received words. This method takes one or more sequences of received words and returns their corresponding estimated message sequences.

        Parameters:
            input: The input sequence(s). Can be either a single sequence whose length is a multiple of $n$, or a multidimensional array where the last dimension is a multiple of $n$.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension contracted from $bn$ to $bk$, where $b$ is a positive integer.
        """
        raise NotImplementedError



from functools import cache
from math import ceil, log2

import numpy.typing as npt
from tqdm import tqdm

from .._util.docs import mkdocstrings
from .._util.information_theory import PMF
from .FixedToVariableCode import FixedToVariableCode
from .util import Word, empty_mapping, extended_probabilities


@mkdocstrings(filters=["!.*"])
class ShannonCode(FixedToVariableCode):
    r"""
    Binary Shannon code. It is a [fixed-to-variable length code](/ref/FixedToVariableCode) in which the length of the codeword $\Enc(u)$ for a source symbol $u \in \mathcal{S}^k$ is given by
    $$
        \ell_u = \left\lceil \log_2 \frac{1}{p_u} \right\rceil,
    $$
    where $p_u$ is the probability of the source symbol $u$. This function implements the lexicographic order assignment as described in [Wikipedia: Shannon–Fano coding](https://en.wikipedia.org/wiki/Shannon%E2%80%93Fano_coding).

    Notes:
        Shannon codes are always [prefix-free](/ref/FixedToVariableCode/#is_prefix_free) (hence [uniquely decodable](/ref/FixedToVariableCode/#is_uniquely_decodable)).

    Parameters:
        pmf: The probability mass function of the source.
        source_block_size: The source block size $k$. The default value is $k = 1$.

    Examples:
        >>> pmf = [0.7, 0.15, 0.15]

        >>> code = komm.ShannonCode(pmf, 1)
        >>> code.enc_mapping
        {(0,): (0,),
         (1,): (1, 0, 0),
         (2,): (1, 0, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.6)

        >>> code = komm.ShannonCode(pmf, 2)
        >>> code.enc_mapping
        {(0, 0): (0, 0),
         (0, 1): (0, 1, 0, 0),
         (0, 2): (0, 1, 0, 1),
         (1, 0): (0, 1, 1, 0),
         (1, 1): (1, 0, 0, 0, 0, 0),
         (1, 2): (1, 0, 0, 0, 0, 1),
         (2, 0): (0, 1, 1, 1),
         (2, 1): (1, 0, 0, 0, 1, 0),
         (2, 2): (1, 0, 0, 0, 1, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.6)
    """

    def __init__(self, pmf: npt.ArrayLike, source_block_size: int = 1):
        self.pmf = PMF(pmf)
        super().__init__(
            source_cardinality=self.pmf.size,
            target_cardinality=2,
            source_block_size=source_block_size,
            enc_mapping=shannon_code(self.pmf, source_block_size),
        )

    def __repr__(self) -> str:
        args = ", ".join([
            f"pmf={self.pmf.tolist()}",
            f"source_block_size={self.source_block_size}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cache
    def is_uniquely_decodable(self) -> bool:
        return True

    @cache
    def is_prefix_free(self) -> bool:
        return True


def next_in_lexicographic_order(word: Word) -> Word:
    word_list = list(word)
    for i in range(len(word_list) - 1, -1, -1):
        if word_list[i] == 0:
            word_list[i] = 1
            break
        word_list[i] = 0
    return tuple(word_list)


def shannon_code(pmf: PMF, source_block_size: int) -> dict[Word, Word]:
    pbar = tqdm(
        desc="Generating Shannon code",
        total=2 * pmf.size**source_block_size,
        delay=2.5,
    )

    enc_mapping = empty_mapping(pmf.size, source_block_size)
    v = ()
    for u, pu in extended_probabilities(pmf, source_block_size, pbar):
        if pu > 0:
            length = max(ceil(log2(1 / pu)), 1)
            v = next_in_lexicographic_order(v) + (0,) * (length - len(v))
            enc_mapping[u] = v
        pbar.update()

    pbar.close()

    return enc_mapping



from dataclasses import dataclass
from functools import cache

import numpy as np
import numpy.typing as npt

from .._util.docs import mkdocstrings
from .SystematicBlockCode import SystematicBlockCode
from .util import extended_parity_submatrix


@mkdocstrings(filters=["!.*"])
@dataclass(eq=False)
class GolayCode(SystematicBlockCode):
    r"""
    Binary Golay code. It is the [linear block code](/ref/BlockCode) with parity submatrix
    $$
    P = \begin{bmatrix}
        1 & 1 & 1 & 1 & 1 & 1 & 1 & 1 & 1 & 1 & 0 \\\\
        0 & 0 & 0 & 0 & 1 & 1 & 1 & 1 & 1 & 1 & 1 \\\\
        0 & 1 & 1 & 1 & 0 & 0 & 0 & 1 & 1 & 1 & 1 \\\\
        1 & 0 & 1 & 1 & 0 & 1 & 1 & 0 & 0 & 1 & 1 \\\\
        1 & 1 & 0 & 1 & 1 & 0 & 1 & 0 & 1 & 0 & 1 \\\\
        1 & 1 & 1 & 0 & 1 & 1 & 0 & 1 & 0 & 0 & 1 \\\\
        0 & 0 & 1 & 1 & 1 & 1 & 0 & 0 & 1 & 0 & 1 \\\\
        0 & 1 & 0 & 1 & 0 & 1 & 1 & 1 & 0 & 0 & 1 \\\\
        0 & 1 & 1 & 0 & 1 & 0 & 1 & 0 & 0 & 1 & 1 \\\\
        1 & 0 & 0 & 1 & 1 & 0 & 0 & 1 & 0 & 1 & 1 \\\\
        1 & 0 & 1 & 0 & 0 & 0 & 1 & 1 & 1 & 0 & 1 \\\\
        1 & 1 & 0 & 0 & 0 & 1 & 0 & 0 & 1 & 1 & 1
    \end{bmatrix}
    $$

    The Golay code has the following parameters:

    - Length: $23$
    - Dimension: $12$
    - Minimum distance: $7$

    Notes:
        - The binary Golay code is a perfect code.

    Attributes:
        extended: If `True`, constructs the code in extended version. The default value is `False`.

    This class represents the code in [systematic form](/ref/SystematicBlockCode), with the information set on the left.

    Examples:
        >>> code = komm.GolayCode()
        >>> (code.length, code.dimension, code.redundancy)
        (23, 12, 11)
        >>> code.minimum_distance()
        7

        >>> code = komm.GolayCode(extended=True)
        >>> (code.length, code.dimension, code.redundancy)
        (24, 12, 12)
        >>> code.minimum_distance()
        8
    """

    extended: bool = False

    def __post_init__(self):
        super().__init__(parity_submatrix=golay_parity_submatrix(self.extended))

    @cache
    def minimum_distance(self) -> int:
        return 8 if self.extended else 7


def golay_parity_submatrix(extended: bool = False) -> npt.NDArray[np.integer]:
    parity_submatrix = np.array([
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
        [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1],
        [0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
        [1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1],
        [1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1],
        [1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1],
        [0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1],
        [0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 1],
        [0, 1, 1, 0, 1, 0, 1, 0, 0, 1, 1],
        [1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1],
        [1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1],
        [1, 1, 0, 0, 0, 1, 0, 0, 1, 1, 1],
    ])
    return extended_parity_submatrix(parity_submatrix) if extended else parity_submatrix



from typing import Any, Literal, TypeVar, cast

import numpy as np
import numpy.typing as npt

from .._util.bit_operations import int_to_bits

T = TypeVar("T", bound=npt.NDArray[np.number])


def cartesian_product(A: T, B: T) -> T:
    r"""
    Computes the Cartesian product of two matrices. See <cite>SA15, eq. (2.2)</cite>, where it is called the 'ordered direct product' and uses a different convention.

    Parameters:
        A: First input matrix, with shape (rA, cA).
        B: Second input matrix, with shape (rB, cB).

    Returns:
        The Cartesian product matrix, with shape (rA * rB, cA + cB)
    """
    rA, cA = A.shape
    rB, cB = B.shape
    product = np.zeros((rA * rB, cA + cB), dtype=A.dtype)
    product = cast(T, product)
    for i, rowA in enumerate(A):
        for j, rowB in enumerate(B):
            product[j * rA + i, :] = np.concatenate((rowA, rowB))
    return product


def labeling_natural(order: int) -> npt.NDArray[np.integer]:
    m = order.bit_length() - 1
    labeling = np.empty((order, m), dtype=int)
    for i in range(order):
        labeling[i, :] = int_to_bits(i, m)[::-1]
    return labeling


def labeling_reflected(order: int) -> npt.NDArray[np.integer]:
    m = order.bit_length() - 1
    labeling = np.empty((order, m), dtype=int)
    for i in range(order):
        labeling[i, :] = int_to_bits(i ^ (i >> 1), m)[::-1]
    return labeling


def labeling_natural_2d(orders: tuple[int, int]) -> npt.NDArray[np.integer]:
    order_I, order_Q = orders
    return cartesian_product(
        labeling_natural(order_I),
        labeling_natural(order_Q),
    )


def labeling_reflected_2d(orders: tuple[int, int]) -> npt.NDArray[np.integer]:
    order_I, order_Q = orders
    return cartesian_product(
        labeling_reflected(order_I),
        labeling_reflected(order_Q),
    )


labelings = {
    "natural": labeling_natural,
    "reflected": labeling_reflected,
    "natural_2d": labeling_natural_2d,
    "reflected_2d": labeling_reflected_2d,
}

LabelingStr = Literal["natural", "reflected", "natural_2d", "reflected_2d"]


def get_labeling(
    labeling: LabelingStr | npt.ArrayLike,
    allowed_labelings: tuple[LabelingStr, ...],
    *args: Any,
    **kwargs: Any,
) -> npt.NDArray[np.integer]:
    if isinstance(labeling, str):
        if labeling not in allowed_labelings:
            raise ValueError(f"if string, 'labeling' must be in {allowed_labelings}")
        return labelings[labeling](*args, **kwargs)
    return np.asarray(labeling, dtype=int)



from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from . import base
from .constellations import constellation_ask
from .labelings import get_labeling


class ASKModulation(base.Modulation[np.complexfloating]):
    r"""
    Amplitude-shift keying (ASK) modulation. It is a complex [modulation scheme](/ref/Modulation) in which the points of the constellation symbols are *uniformly arranged* in a ray. More precisely, the $i$-th constellation symbol is given by
    $$
        x_i = iA \exp(\mathrm{j}\phi), \quad i \in [0 : M),
    $$
    where $M$ is the *order* (a power of $2$), $A$ is the *base amplitude*, and $\phi$ is the *phase offset* of the modulation.

    Parameters:
        order: The order $M$ of the modulation. It must be a power of $2$.

        base_amplitude: The base amplitude $A$ of the constellation. The default value is `1.0`.

        phase_offset: The phase offset $\phi$ of the constellation. The default value is `0.0`.

        labeling: The binary labeling of the modulation. Can be specified either as a 2D-array of integers (see [base class](/ref/Modulation) for details), or as a string. In the latter case, the string must be either `'natural'` or `'reflected'`. The default value is `'reflected'`, corresponding to the Gray labeling.

    Examples:
        1. The $4$-ASK modulation with base amplitude $A = 1$, phase offset $\phi = 0$, and Gray labeling is depicted below.
            <figure markdown>
            ![4-ASK modulation with Gray labeling.](/figures/ask_4_gray.svg)
            </figure>

                >>> ask = komm.ASKModulation(4)
                >>> ask.constellation
                array([0.+0.j, 1.+0.j, 2.+0.j, 3.+0.j])
                >>> ask.labeling
                array([[0, 0],
                       [0, 1],
                       [1, 1],
                       [1, 0]])

        2. The $4$-ASK modulation with base amplitude $A = 2\sqrt{2}$, phase offset $\phi = \pi/4$, and natural labeling is depicted below.
            <figure markdown>
            ![4-ASK modulation with natural labeling.](/figures/ask_4_natural.svg)
            </figure>

                >>> ask = komm.ASKModulation(
                ...     order=4,
                ...     base_amplitude=2*np.sqrt(2),
                ...     phase_offset=np.pi/4,
                ...     labeling='natural',
                ... )
                >>> ask.constellation
                array([0.+0.j, 2.+2.j, 4.+4.j, 6.+6.j])
                >>> ask.labeling
                array([[0, 0],
                       [0, 1],
                       [1, 0],
                       [1, 1]])
    """

    def __init__(
        self,
        order: int,
        base_amplitude: float = 1.0,
        phase_offset: float = 0.0,
        labeling: Literal["natural", "reflected"] | npt.ArrayLike = "reflected",
    ) -> None:
        self._order = order
        self._base_amplitude = base_amplitude
        self._phase_offset = phase_offset
        self._labeling = labeling
        super()._validate_parameters()

    def __repr__(self) -> str:
        if isinstance(self._labeling, str):
            labeling_repr = repr(self._labeling)
        else:
            labeling_repr = self.labeling.tolist()
        args = ", ".join([
            f"order={self._order}",
            f"base_amplitude={self._base_amplitude}",
            f"phase_offset={self._phase_offset}",
            f"labeling={labeling_repr}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def constellation(self) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.constellation
            array([0.+0.j, 1.+0.j, 2.+0.j, 3.+0.j])
        """
        return constellation_ask(self._order, self._base_amplitude, self._phase_offset)

    @cached_property
    def labeling(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.labeling
            array([[0, 0],
                   [0, 1],
                   [1, 1],
                   [1, 0]])
        """
        return get_labeling(self._labeling, ("natural", "reflected"), self._order)

    @cached_property
    def inverse_labeling(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.inverse_labeling
            {(0, 0): 0, (0, 1): 1, (1, 1): 2, (1, 0): 3}
        """
        return super().inverse_labeling

    @cached_property
    def order(self) -> int:
        r"""
        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.order
            4
        """
        return self._order

    @cached_property
    def bits_per_symbol(self) -> int:
        r"""
        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.bits_per_symbol
            2
        """
        return super().bits_per_symbol

    @cached_property
    def energy_per_symbol(self) -> float:
        r"""
        For the ASK, it is given by
        $$
            E_\mathrm{s} = \frac{A^2}{6} (M - 1) (2M - 1).
        $$

        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.energy_per_symbol
            3.5
        """
        return self._base_amplitude**2 * (self._order - 1) * (2 * self._order - 1) / 6

    @cached_property
    def energy_per_bit(self) -> float:
        r"""
        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.energy_per_bit
            1.75
        """
        return super().energy_per_bit

    @cached_property
    def symbol_mean(self) -> complex:
        r"""
        For the ASK, it is given by
        $$
            \mu_\mathrm{s} = \frac{A}{2} (M-1) \exp(\mathrm{j}\phi).
        $$

        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.symbol_mean
            (1.5+0j)
        """
        M, A, phi = self._order, self._base_amplitude, self._phase_offset
        return complex(0.5 * A * (M - 1) * np.exp(1j * phi))

    @cached_property
    def minimum_distance(self) -> float:
        r"""
        For the ASK, it is given by
        $$
            d_\mathrm{min} = A.
        $$

        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.minimum_distance
            1.0
        """
        return self._base_amplitude

    def modulate(self, input: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> ask = komm.ASKModulation(4)
            >>> ask.modulate([0, 0, 1, 1, 0, 0, 0, 1])
            array([0.+0.j, 2.+0.j, 0.+0.j, 1.+0.j])
        """
        return super().modulate(input)

    def demodulate_hard(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        M, A, phi = self._order, self._base_amplitude, self._phase_offset
        input = np.real(np.multiply(input, np.exp(-1j * phi))) / A
        indices = np.clip(np.round(input).astype(int), 0, M - 1)
        hard_bits = np.reshape(self.labeling[indices], shape=-1)
        return hard_bits

    def demodulate_soft(
        self, input: npt.ArrayLike, snr: float = 1.0
    ) -> npt.NDArray[np.floating]:
        return super().demodulate_soft(input, snr)



from abc import ABC, abstractmethod
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .._util.information_theory import LogBase


class DiscreteMemorylessChannel(ABC):
    @cached_property
    @abstractmethod
    def input_cardinality(self) -> int:
        r"""
        The channel input cardinality $|\mathcal{X}|$.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def output_cardinality(self) -> int:
        r"""
        The channel output cardinality $|\mathcal{Y}|$.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def transition_matrix(self) -> npt.NDArray[np.floating]:
        r"""
        The channel transition probability matrix $p_{Y \mid X}$.
        """
        raise NotImplementedError

    @abstractmethod
    def mutual_information(
        self, input_pmf: npt.ArrayLike, base: LogBase = 2.0
    ) -> float:
        r"""Returns the mutual information $\mathrm{I}(X ; Y)$ between the input $X$ and the output $Y$ of the channel.

        Parameters:
            input_pmf: The probability mass function $p_X$ of the channel input $X$. It must be a valid pmf, that is, all of its values must be non-negative and sum up to $1$.

            base: The base of the logarithm to be used. It must be a positive float or the string `'e'`. The default value is `2.0`.

        Returns:
            The mutual information $\mathrm{I}(X ; Y)$ between the input $X$ and the output $Y$.
        """
        raise NotImplementedError

    @abstractmethod
    def capacity(self, base: LogBase = 2.0) -> float:
        r"""
        Returns the channel capacity $C$.

        Parameters:
            base: The base of the logarithm to be used. It must be a positive float or the string `'e'`. The default value is `2.0`.

        Returns:
            The channel capacity $C$.
        """
        raise NotImplementedError

    @abstractmethod
    def __call__(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Transmits the input sequence through the channel and returns the output sequence.

        Parameters:
            input: The input sequence.

        Returns:
            output: The output sequence.
        """
        raise NotImplementedError



from dataclasses import dataclass
from functools import cache, cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from .._error_control_block import base
from .._util.bit_operations import bits_to_int, int_to_bits
from .._util.decorators import blockwise, vectorize
from .._util.matrices import null_matrix, pseudo_inverse
from . import terminations
from .ConvolutionalCode import ConvolutionalCode

TerminationMode = Literal["direct-truncation", "zero-termination", "tail-biting"]


@dataclass(eq=False)
class TerminatedConvolutionalCode(base.BlockCode):
    r"""
    Terminated convolutional code. It is a [linear block code](/ref/BlockCode) obtained by terminating a $(n_0, k_0)$ [convolutional code](/ref/ConvolutionalCode). A total of $h$ information blocks (each containing $k_0$ information bits) is encoded. The dimension of the resulting block code is thus $k = h k_0$; its length depends on the termination mode employed. There are three possible termination modes:

    - **Direct truncation**. The encoder always starts at state $0$, and its output ends immediately after the last information block. The encoder may not necessarily end in state $0$. The resulting block code will have length $n = h n_0$.

    - **Zero termination**. The encoder always starts and ends at state $0$. To achieve this, a sequence of $k \mu$ tail bits is appended to the information bits, where $\mu$ is the memory order of the convolutional code. The resulting block code will have length $n = (h + \mu) n_0$.

    - **Tail-biting**. The encoder always starts and ends at the same state. To achieve this, the initial state of the encoder is chosen as a function of the information bits. The resulting block code will have length $n = h n_0$.

    For more details, see <cite>LC04, Sec. 12.7</cite> and <cite>WBR01</cite>.

    Attributes:
        convolutional_code: The convolutional code to be terminated.

        num_blocks: The number $h$ of information blocks.

        mode: The termination mode. It must be one of `'direct-truncation'` | `'zero-termination'` | `'tail-biting'`. The default value is `'zero-termination'`.

    Examples:
        >>> convolutional_code = komm.ConvolutionalCode([[0b1, 0b11]])

        >>> code = komm.TerminatedConvolutionalCode(convolutional_code, num_blocks=3, mode='direct-truncation')
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[1, 1, 0, 1, 0, 0],
               [0, 0, 1, 1, 0, 1],
               [0, 0, 0, 0, 1, 1]])
        >>> code.minimum_distance()
        2

        >>> code = komm.TerminatedConvolutionalCode(convolutional_code, num_blocks=3, mode='zero-termination')
        >>> (code.length, code.dimension, code.redundancy)
        (8, 3, 5)
        >>> code.generator_matrix
        array([[1, 1, 0, 1, 0, 0, 0, 0],
               [0, 0, 1, 1, 0, 1, 0, 0],
               [0, 0, 0, 0, 1, 1, 0, 1]])
        >>> code.minimum_distance()
        3

        >>> code = komm.TerminatedConvolutionalCode(convolutional_code, num_blocks=3, mode='tail-biting')
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[1, 1, 0, 1, 0, 0],
               [0, 0, 1, 1, 0, 1],
               [0, 1, 0, 0, 1, 1]])
        >>> code.minimum_distance()
        3
    """

    convolutional_code: ConvolutionalCode
    num_blocks: int
    mode: TerminationMode = "zero-termination"

    def __post_init__(self):
        if not self.mode in TerminationMode.__args__:
            raise ValueError(
                f"mode '{self.mode}' is unknown\n"
                f"supported termination modes: {set(TerminationMode.__args__)}"
            )

    @cached_property
    def _strategy(self) -> terminations.TerminationStrategy:
        return {
            "direct-truncation": terminations.DirectTruncation,
            "zero-termination": terminations.ZeroTermination,
            "tail-biting": terminations.TailBiting,
        }[self.mode](self.convolutional_code, self.num_blocks)

    @cached_property
    def length(self) -> int:
        return self._strategy.codeword_length()

    @cached_property
    def dimension(self) -> int:
        return self.num_blocks * self.convolutional_code.num_input_bits

    @cached_property
    def redundancy(self) -> int:
        return self.length - self.dimension

    @cached_property
    def rate(self) -> float:
        return super().rate

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.integer]:
        return self._strategy.generator_matrix(self)

    @cached_property
    def generator_matrix_right_inverse(self) -> npt.NDArray[np.integer]:
        return pseudo_inverse(self.generator_matrix)

    @cached_property
    def check_matrix(self) -> npt.NDArray[np.integer]:
        return null_matrix(self.generator_matrix)

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        k0 = self.convolutional_code.num_input_bits
        n0 = self.convolutional_code.num_output_bits
        fsm = self.convolutional_code.finite_state_machine()

        @blockwise(self.dimension)
        @vectorize
        def encode(u: npt.NDArray[np.integer]) -> npt.NDArray[np.integer]:
            u = self._strategy.pre_process_input(u)
            input_sequence = bits_to_int(u.reshape(-1, k0))
            initial_state = self._strategy.initial_state(input_sequence)
            output_sequence, _ = fsm.process(input_sequence, initial_state)
            v = int_to_bits(output_sequence, width=n0).ravel()
            return v

        return encode(input)

    def project_word(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return super().project_word(input)

    def inverse_encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return super().inverse_encode(input)

    def check(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return super().check(input)

    @cache
    def codewords(self) -> npt.NDArray[np.integer]:
        return super().codewords()

    @cache
    def codeword_weight_distribution(self) -> npt.NDArray[np.integer]:
        return super().codeword_weight_distribution()

    @cache
    def minimum_distance(self) -> int:
        return super().minimum_distance()

    @cache
    def coset_leaders(self) -> npt.NDArray[np.integer]:
        return super().coset_leaders()

    @cache
    def coset_leader_weight_distribution(self) -> npt.NDArray[np.integer]:
        return super().coset_leader_weight_distribution()

    @cache
    def packing_radius(self) -> int:
        return super().packing_radius()

    @cache
    def covering_radius(self) -> int:
        return super().covering_radius()



from functools import cached_property

import numpy as np
import numpy.typing as npt

from .._util.correlation import autocorrelation, cyclic_autocorrelation


class BinarySequence:
    r"""
    General binary sequence. It may be represented either in *bit format*, denoted by $b[n]$, with elements in the set $\\{ 0, 1 \\}$, or in *polar format*, denoted by $x[n]$, with elements in the set $\\{ \pm 1 \\}$. The correspondences $0 \mapsto +1$ and $1 \mapsto -1$ from bit format to polar format is assumed.

    The constructor expects either the bit sequence or the polar sequence.

    Parameters:
        bit_sequence: The binary sequence in bit format, $b[n] \in \\{ 0, 1 \\}$.
        polar_sequence: The binary sequence in polar format, $x[n] \in \\{ \pm 1 \\}$.

    Examples:
        >>> seq = komm.BinarySequence(bit_sequence=[0, 1, 1, 0])
        >>> seq.bit_sequence
        array([0, 1, 1, 0])
        >>> seq.polar_sequence
        array([ 1, -1, -1,  1])

        >>> seq = komm.BinarySequence(polar_sequence=[1, -1, -1, 1])
        >>> seq.bit_sequence
        array([0, 1, 1, 0])
        >>> seq.polar_sequence
        array([ 1, -1, -1,  1])
    """

    def __init__(
        self,
        bit_sequence: npt.ArrayLike | None = None,
        polar_sequence: npt.ArrayLike | None = None,
    ) -> None:
        if bit_sequence is not None and polar_sequence is None:
            self.bit_sequence = np.asarray(bit_sequence, dtype=int)
            self.polar_sequence = (-1) ** self.bit_sequence
        elif polar_sequence is not None and bit_sequence is None:
            self.polar_sequence = np.asarray(polar_sequence, dtype=int)
            self.bit_sequence = 1 * (self.polar_sequence < 0)
        else:
            raise ValueError("either specify 'bit_sequence' or 'polar_sequence'")

    def __repr__(self) -> str:
        args = f"bit_sequence={self.bit_sequence.tolist()}"
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def length(self) -> int:
        r"""
        The length (or period) $L$ of the binary sequence.
        """
        return self.bit_sequence.size

    def autocorrelation(
        self, shifts: npt.ArrayLike | None = None, normalized: bool = False
    ) -> npt.NDArray[np.floating]:
        r"""
        Returns the autocorrelation $R[\ell]$ of the binary sequence in polar format. See [`komm.autocorrelation`](/ref/autocorrelation) for more details.

        Parameters:
            shifts: See the corresponding parameter in [`komm.autocorrelation`](/ref/autocorrelation).
            normalized: See the corresponding parameter in [`komm.autocorrelation`](/ref/autocorrelation).

        Returns:
            The autocorrelation $R[\ell]$ of the binary sequence.

        Examples:
            >>> seq = komm.BinarySequence(bit_sequence=[0, 1, 1, 0])
            >>> seq.autocorrelation()
            array([ 4, -1, -2,  1])
        """
        return autocorrelation(
            self.polar_sequence, shifts=shifts, normalized=normalized
        )

    def cyclic_autocorrelation(
        self, shifts: npt.ArrayLike | None = None, normalized: bool = False
    ) -> npt.NDArray[np.floating]:
        r"""
        Returns the cyclic autocorrelation $\tilde{R}[\ell]$ of the binary sequence in polar format. See [`komm.cyclic_autocorrelation`](/ref/cyclic_autocorrelation) for more details.

        Parameters:
            shifts: See the corresponding parameter in [`komm.cyclic_autocorrelation`](/ref/cyclic_autocorrelation).
            normalized: See the corresponding parameter in [`komm.cyclic_autocorrelation`](/ref/cyclic_autocorrelation).

        Returns:
            The cyclic autocorrelation $\tilde{R}[\ell]$ of the binary sequence.

        Examples:
            >>> seq = komm.BinarySequence(bit_sequence=[0, 1, 1, 0])
            >>> seq.cyclic_autocorrelation()
            array([ 4,  0, -4,  0])
        """
        return cyclic_autocorrelation(
            self.polar_sequence, shifts=shifts, normalized=normalized
        )



import numpy as np
import numpy.typing as npt


def constellation_pam(
    order: int,
    base_amplitude: float,
) -> npt.NDArray[np.floating]:
    return base_amplitude * np.arange(-order + 1, order, step=2, dtype=int)


def constellation_qam(
    orders: tuple[int, int],
    base_amplitudes: tuple[float, float],
    phase_offset: float,
) -> npt.NDArray[np.complexfloating]:
    order_I, order_Q = orders
    base_amplitude_I, base_amplitude_Q = base_amplitudes
    constellation_I = constellation_pam(order_I, base_amplitude_I)
    constellation_Q = constellation_pam(order_Q, base_amplitude_Q)
    constellation = (constellation_I + 1j * constellation_Q[np.newaxis].T).flatten()
    constellation *= np.exp(1j * phase_offset)
    return constellation


def constellation_ask(
    order: int,
    base_amplitude: float,
    phase_offset: float,
) -> npt.NDArray[np.complexfloating]:
    return base_amplitude * np.arange(order, dtype=int) * np.exp(1j * phase_offset)


def constellation_psk(
    order: int,
    amplitude: float,
    phase_offset: float,
) -> npt.NDArray[np.complexfloating]:
    return (
        amplitude
        * np.exp(2j * np.pi * np.arange(order) / order)
        * np.exp(1j * phase_offset)
    )


def constellation_apsk(
    orders: tuple[int, ...],
    amplitudes: tuple[float, ...],
    phase_offsets: tuple[float, ...],
) -> npt.NDArray[np.complexfloating]:
    return np.concatenate([
        constellation_psk(order, amplitude, phase_offset)
        for order, amplitude, phase_offset in zip(orders, amplitudes, phase_offsets)
    ])



from dataclasses import dataclass
from functools import cache

from .._util.docs import mkdocstrings
from .HammingCode import hamming_parity_submatrix
from .SystematicBlockCode import SystematicBlockCode


@mkdocstrings(filters=["!.*"])
@dataclass(eq=False)
class SimplexCode(SystematicBlockCode):
    r"""
    Simplex (maximum-length) code. For a given parameter $\kappa \geq 2$, it is the [linear block code](/ref/BlockCode) with generator matrix whose columns are all the $2^\kappa - 1$ nonzero binary $\kappa$-tuples. The simplex code (also known as maximum-length code) has the following parameters:

    - Length: $n = 2^\kappa - 1$
    - Dimension: $k = \kappa$
    - Redundancy: $m = 2^\kappa - \kappa - 1$
    - Minimum distance: $d = 2^{\kappa - 1}$

    In its extended version, the simplex code has the following parameters:

    - Length: $n = 2^\kappa$
    - Dimension: $k = \kappa + 1$
    - Redundancy: $m = 2^\kappa - \kappa - 1$
    - Minimum distance: $d = 2^{\kappa - 1}$

    Notes:
        - For $\kappa = 2$ it reduces to the [single parity-check code](/ref/SingleParityCheckCode) of length $3$.
        - Its dual is the [Hamming code](/ref/HammingCode).
        - Simplex codes are constant-weight codes.

    Attributes:
        kappa: The parameter $\kappa$ of the code. Must satisfy $\kappa \geq 2$.
        extended: Whether to use the extended version of the Simplex code. Default is `False`.

    This class represents the code in [systematic form](/ref/SystematicBlockCode), with the information set on the left.

    Examples:
        >>> code = komm.SimplexCode(3)
        >>> (code.length, code.dimension, code.redundancy)
        (7, 3, 4)
        >>> code.generator_matrix
        array([[1, 0, 0, 1, 1, 0, 1],
               [0, 1, 0, 1, 0, 1, 1],
               [0, 0, 1, 0, 1, 1, 1]])
        >>> code.check_matrix
        array([[1, 1, 0, 1, 0, 0, 0],
               [1, 0, 1, 0, 1, 0, 0],
               [0, 1, 1, 0, 0, 1, 0],
               [1, 1, 1, 0, 0, 0, 1]])
        >>> code.minimum_distance()
        4

        >>> code = komm.SimplexCode(3, extended=True)
        >>> (code.length, code.dimension, code.redundancy)
        (8, 4, 4)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1, 1, 0, 1],
               [0, 1, 0, 0, 1, 0, 1, 1],
               [0, 0, 1, 0, 0, 1, 1, 1],
               [0, 0, 0, 1, 1, 1, 1, 0]])
        >>> code.check_matrix
        array([[1, 1, 0, 1, 1, 0, 0, 0],
               [1, 0, 1, 1, 0, 1, 0, 0],
               [0, 1, 1, 1, 0, 0, 1, 0],
               [1, 1, 1, 0, 0, 0, 0, 1]])
        >>> code.minimum_distance()
        4
    """

    kappa: int
    extended: bool = False

    def __post_init__(self):
        if not self.kappa >= 2:
            raise ValueError("'kappa' must be at least 2")
        super().__init__(
            parity_submatrix=hamming_parity_submatrix(self.kappa, self.extended).T
        )

    @cache
    def minimum_distance(self) -> int:
        return 2 ** (self.kappa - 1)



from functools import cache, cached_property

import numpy as np
import numpy.typing as npt

from .._util.matrices import null_matrix, pseudo_inverse, rref
from . import base


class BlockCode(base.BlockCode):
    r"""
    General binary linear block code. It is characterized by its *generator matrix* $G \in \mathbb{B}^{k \times n}$, and by its *check matrix* $H \in \mathbb{B}^{m \times n}$, which are related by $G H^\transpose = 0$. The parameters $n$, $k$, and $m$ are called the code *length*, *dimension*, and *redundancy*, respectively, and are related by $k + m = n$. For more details, see <cite>LC04, Ch. 3</cite>.

    The constructor expects either the generator matrix or the check matrix.

    Attributes:
        generator_matrix: The generator matrix $G$ of the code, which is a $k \times n$ binary matrix.

        check_matrix: The check matrix $H$ of the code, which is a $m \times n$ binary matrix.

    Examples:
        >>> code = komm.BlockCode(generator_matrix=[
        ...     [1, 0, 0, 1, 1],
        ...     [0, 1, 1, 1, 0],
        ... ])
        >>> (code.length, code.dimension, code.redundancy)
        (5, 2, 3)
        >>> code.generator_matrix
        array([[1, 0, 0, 1, 1],
               [0, 1, 1, 1, 0]])
        >>> code.check_matrix
        array([[0, 1, 1, 0, 0],
               [1, 1, 0, 1, 0],
               [1, 0, 0, 0, 1]])

        >>> code = komm.BlockCode(check_matrix=[
        ...     [0, 1, 1, 0, 0],
        ...     [1, 1, 0, 1, 0],
        ...     [1, 0, 0, 0, 1],
        ... ])
        >>> (code.length, code.dimension, code.redundancy)
        (5, 2, 3)
        >>> code.generator_matrix
        array([[1, 0, 0, 1, 1],
               [0, 1, 1, 1, 0]])
        >>> code.check_matrix
        array([[0, 1, 1, 0, 0],
               [1, 1, 0, 1, 0],
               [1, 0, 0, 0, 1]])
    """

    def __init__(
        self,
        generator_matrix: npt.ArrayLike | None = None,
        check_matrix: npt.ArrayLike | None = None,
    ):
        if generator_matrix is None and check_matrix is None:
            raise ValueError(
                "either 'generator_matrix' or 'check_matrix' must be provided"
            )
        if generator_matrix is not None and check_matrix is not None:
            raise ValueError(
                "only one of 'generator_matrix' or 'check_matrix' must be provided"
            )
        if generator_matrix is not None:
            self._generator_matrix = np.asarray(generator_matrix)
            self._check_matrix = None
            self._dimension, self._length = self._generator_matrix.shape
            self._redundancy = self._length - self._dimension
        else:  # check_matrix is not None
            self._generator_matrix = None
            self._check_matrix = np.asarray(check_matrix)
            self._redundancy, self._length = self._check_matrix.shape
            self._dimension = self._length - self._redundancy

    def __repr__(self) -> str:
        if self._generator_matrix is not None:
            args = f"generator_matrix={self.generator_matrix.tolist()}"
        else:  # self._check_matrix is not None
            args = f"check_matrix={self.check_matrix.tolist()}"
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def length(self) -> int:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.length
            5
        """
        return self._length

    @cached_property
    def dimension(self) -> int:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.dimension
            2
        """
        return self._dimension

    @cached_property
    def redundancy(self) -> int:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.redundancy
            3
        """
        return self._redundancy

    @cached_property
    def rate(self) -> float:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.rate
            0.4
        """
        return super().rate

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.BlockCode(check_matrix=[
            ...     [0, 1, 1, 0, 0],
            ...     [1, 1, 0, 1, 0],
            ...     [1, 0, 0, 0, 1],
            ... ])
            >>> code.generator_matrix
            array([[1, 0, 0, 1, 1],
                   [0, 1, 1, 1, 0]])
        """
        if self._generator_matrix is not None:
            return self._generator_matrix
        return rref(null_matrix(self.check_matrix))

    @cached_property
    def generator_matrix_right_inverse(self) -> npt.NDArray[np.integer]:
        return pseudo_inverse(self.generator_matrix)

    @cached_property
    def check_matrix(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.check_matrix
            array([[0, 1, 1, 0, 0],
                   [1, 1, 0, 1, 0],
                   [1, 0, 0, 0, 1]])
        """
        if self._check_matrix is not None:
            return self._check_matrix
        return null_matrix(self.generator_matrix)

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.encode([0, 0])  # Sequence with single message
            array([0, 0, 0, 0, 0])
            >>> code.encode([0, 0, 1, 1])  # Sequence with two messages
            array([0, 0, 0, 0, 0, 1, 1, 1, 0, 1])
            >>> code.encode([[0, 0],  # 2D array of single messages
            ...              [1, 1]])
            array([[0, 0, 0, 0, 0],
                   [1, 1, 1, 0, 1]])
            >>> code.encode([[0, 0, 1, 1],  # 2D array of two messages
            ...              [1, 1, 1, 0]])
            array([[0, 0, 0, 0, 0, 1, 1, 1, 0, 1],
                   [1, 1, 1, 0, 1, 1, 0, 0, 1, 1]])
        """
        return super().encode(input)

    def project_word(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return super().project_word(input)

    def inverse_encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.inverse_encode([0, 0, 0, 0, 0])  # Sequence with single codeword
            array([0, 0])
            >>> code.inverse_encode([0, 0, 0, 0, 0, 1, 1, 1, 0, 1])  # Sequence with two codewords
            array([0, 0, 1, 1])
            >>> code.inverse_encode([[0, 0, 0, 0, 0],  # 2D array of single codewords
            ...                      [1, 1, 1, 0, 1]])
            array([[0, 0],
                   [1, 1]])
            >>> code.inverse_encode([[0, 0, 0, 0, 0, 1, 1, 1, 0, 1],  # 2D array of two codewords
            ...                      [1, 1, 1, 0, 1, 1, 0, 0, 1, 1]])
            array([[0, 0, 1, 1],
                   [1, 1, 1, 0]])
        """
        return super().inverse_encode(input)

    def check(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.check([1, 1, 1, 0, 1])  # Sequence with single received word
            array([0, 0, 0])
            >>> code.check([1, 1, 1, 0, 1, 1, 1, 1, 1, 1])  # Sequence with two received words
            array([0, 0, 0, 0, 1, 0])
            >>> code.check([[1, 1, 1, 0, 1],  # 2D array of single received words
            ...             [1, 1, 1, 1, 1]])
            array([[0, 0, 0],
                   [0, 1, 0]])
            >>> code.check([[1, 1, 1, 0, 1, 1, 1, 1, 1, 1],  # 2D array of two received words
            ...             [1, 1, 1, 1, 1, 0, 0, 0, 1, 1]])
            array([[0, 0, 0, 0, 1, 0],
                   [0, 1, 0, 0, 1, 1]])
        """
        return super().check(input)

    @cache
    def codewords(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.codewords()
            array([[0, 0, 0, 0, 0],
                   [1, 0, 0, 1, 1],
                   [0, 1, 1, 1, 0],
                   [1, 1, 1, 0, 1]])
        """
        return super().codewords()

    @cache
    def codeword_weight_distribution(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.codeword_weight_distribution()
            array([1, 0, 0, 2, 1, 0])
        """
        return super().codeword_weight_distribution()

    @cache
    def minimum_distance(self) -> int:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.minimum_distance()
            3
        """
        return super().minimum_distance()

    @cache
    def coset_leaders(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.coset_leaders()
            array([[0, 0, 0, 0, 0],
                   [0, 0, 1, 0, 0],
                   [0, 0, 0, 1, 0],
                   [0, 1, 0, 0, 0],
                   [0, 0, 0, 0, 1],
                   [1, 1, 0, 0, 0],
                   [1, 0, 0, 0, 0],
                   [1, 0, 1, 0, 0]])
        """
        return super().coset_leaders()

    @cache
    def coset_leader_weight_distribution(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.coset_leader_weight_distribution()
            array([1, 5, 2, 0, 0, 0])
        """
        return super().coset_leader_weight_distribution()

    @cache
    def packing_radius(self) -> int:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.packing_radius()
            1
        """
        return super().packing_radius()

    @cache
    def covering_radius(self) -> int:
        r"""
        Examples:
            >>> code = komm.BlockCode(generator_matrix=[
            ...     [1, 0, 0, 1, 1],
            ...     [0, 1, 1, 1, 0],
            ... ])
            >>> code.covering_radius()
            2
        """
        return super().covering_radius()



from dataclasses import dataclass
from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from .._finite_state_machine.FiniteStateMachine import MetricMemory
from .._util.bit_operations import int_to_bits
from .ConvolutionalCode import ConvolutionalCode


@dataclass
class ViterbiStreamDecoder:
    r"""
    Convolutional stream decoder using Viterbi algorithm. Decode a (hard or soft) bit stream given a [convolutional code](/ref/ConvolutionalCode), assuming a traceback length (path memory) of $\tau$. At time $t$, the decoder chooses the path survivor with best metric at time $t - \tau$ and outputs the corresponding information bits. The output stream has a delay equal to $k \tau$, where $k$ is the number of input bits of the convolutional code. As a rule of thumb, the traceback length is chosen as $\tau = 5\mu$, where $\mu$ is the memory order of the convolutional code.

    Attributes:
        convolutional_code: The convolutional code.
        traceback_length: The traceback length (path memory) $\tau$ of the decoder.
        state: The current state of the decoder. The default value is `0`.
        input_type: The type of the input sequence, either `hard` or `soft`. The default value is `hard`.
    """

    convolutional_code: ConvolutionalCode
    traceback_length: int
    state: int = 0
    input_type: Literal["hard", "soft"] = "hard"

    def __post_init__(self):
        fsm = self.convolutional_code.finite_state_machine()
        num_states, traceback_length = fsm.num_states, self.traceback_length
        self.memory: MetricMemory = {
            "paths": np.zeros((num_states, traceback_length + 1), dtype=int),
            "metrics": np.full((num_states, traceback_length + 1), fill_value=np.inf),
        }
        self.memory["metrics"][self.state, -1] = 0.0

    @cached_property
    def cache_bit(self) -> npt.NDArray[np.integer]:
        n = self.convolutional_code.num_output_bits
        return np.array([int_to_bits(y, width=n) for y in range(2**n)])

    def metric_function(self, y: int, z: npt.ArrayLike) -> float:
        if self.input_type == "hard":
            return np.count_nonzero(self.cache_bit[y] != z)
        else:  # self.input_type == "soft"
            return np.dot(self.cache_bit[y], z)

    def __call__(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Parameters:
            input: The (hard or soft) bit sequence to be decoded.

        Returns:
            output: The decoded bit sequence.

        Examples:
                >>> convolutional_code = komm.ConvolutionalCode([[0o7, 0o5]])
                >>> decoder = komm.ViterbiStreamDecoder(convolutional_code, traceback_length=10)
                >>> decoder([1, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 1, 1])
                array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
                >>> decoder(np.zeros(2*10, dtype=int))
                array([1, 0, 1, 1, 1, 0, 1, 1, 0, 0])
        """
        input = np.asarray(input)
        n = self.convolutional_code.num_output_bits
        k = self.convolutional_code.num_input_bits
        fsm = self.convolutional_code.finite_state_machine()
        input_sequence_hat = fsm.viterbi_streaming(
            observed_sequence=input.reshape(-1, n),
            metric_function=self.metric_function,
            memory=self.memory,
        )
        output = int_to_bits(input_sequence_hat, width=k).ravel()
        return output



from .FanoCode import FanoCode
from .FixedToVariableCode import FixedToVariableCode
from .HuffmanCode import HuffmanCode
from .LempelZiv78Code import LempelZiv78Code
from .LempelZivWelchCode import LempelZivWelchCode
from .ShannonCode import ShannonCode
from .TunstallCode import TunstallCode
from .VariableToFixedCode import VariableToFixedCode

__all__ = [
    "FanoCode",
    "FixedToVariableCode",
    "HuffmanCode",
    "LempelZiv78Code",
    "LempelZivWelchCode",
    "ShannonCode",
    "TunstallCode",
    "VariableToFixedCode",
]



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from . import base


@dataclass
class GaussianPulse(base.Pulse):
    r"""
    Gaussian pulse. It is a [pulse](/ref/Pulse) with waveform given by
    $$
        h(t) = \mathrm{e}^{-\frac{1}{2} (2 \pi \bar{B} t)^2}
    $$
    where the $\bar{B} = B / \sqrt{\ln 2}$, and $B$ is the _half-power bandwidth_ of the filter. Its spectrum is given by
    $$
        \hat{h}(f) = \frac{1}{\sqrt{2 \pi} \bar{B}} \mathrm{e}^{-\frac{1}{2} (f / \bar{B})^2}.
    $$

    The Gaussian pulse is depicted below for $B = 0.5$, and for $B = 1$.

    <div class="centered" markdown>
      <span>
        ![Gaussian pulse with half-power bandwidth of 0.5.](/figures/pulse_gaussian_50.svg)
      </span>
      <span>
        ![Gaussian pulse with half-power bandwidth of 1.](/figures/pulse_gaussian_100.svg)
      </span>
    </div>

    Attributes:
        half_power_bandwidth: The half-power bandwidth $B$ of the pulse. The default value is `1.0`.

    Examples:
        >>> pulse = komm.GaussianPulse(half_power_bandwidth=0.25)
        >>> pulse.waveform([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75]).round(4)
        array([0.3675, 0.6408, 0.8947, 1.    , 0.8947, 0.6408, 0.3675])
        >>> pulse.spectrum([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75]).round(4)
        array([0.0587, 0.3321, 0.9394, 1.3286, 0.9394, 0.3321, 0.0587])
    """

    half_power_bandwidth: float = 1.0

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        t = np.asarray(t)
        b_bar = self.half_power_bandwidth / np.sqrt(np.log(2))
        return np.exp(-0.5 * (2 * np.pi * b_bar * t) ** 2)

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        f = np.asarray(f)
        b_bar = self.half_power_bandwidth / np.sqrt(np.log(2))
        return 1 / (np.sqrt(2 * np.pi) * b_bar) * np.exp(-0.5 * (f / b_bar) ** 2)

    @cached_property
    def support(self) -> tuple[float, float]:
        return (-np.inf, np.inf)



from dataclasses import dataclass
from functools import cache

import numpy as np
import numpy.typing as npt

from .._util.docs import mkdocstrings
from .BlockCode import BlockCode


@mkdocstrings(filters=["!.*"])
@dataclass(eq=False)
class CordaroWagnerCode(BlockCode):
    r"""
    Cordaro–Wagner code. For a given length $n \geq 2$, it is the [linear block code](/ref/BlockCode) with dimension $k = 2$ which is optimum for the [BSC](/ref/BinarySymmetricChannel) with sufficiently small crossover probability. For more details, see <cite>CW67</cite>.

    - Length: $n$
    - Dimension: $k = 2$
    - Redundancy: $m = n - 2$
    - Minimum distance: $d = \left\lceil 2n / 3 \right\rceil - 1$

    Attributes:
        n: The length $n$ of the code. Must satisfy $n \geq 2$.

    Examples:
        >>> code = komm.CordaroWagnerCode(11)
        >>> (code.length, code.dimension, code.redundancy)
        (11, 2, 9)
        >>> code.generator_matrix
        array([[1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1],
               [0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1]])
        >>> code.minimum_distance()
        7
        >>> code.codeword_weight_distribution()
        array([1, 0, 0, 0, 0, 0, 0, 2, 1, 0, 0, 0])
        >>> code.coset_leader_weight_distribution()
        array([  1,  11,  55, 165, 226,  54,   0,   0,   0,   0,   0,   0])
    """

    n: int

    def __post_init__(self):
        if not self.n >= 2:
            raise ValueError("n must be at least 2")
        super().__init__(generator_matrix=cordaro_wagner_generator_matrix(self.n))

    @cache
    def minimum_distance(self) -> int:
        return int(np.ceil(2 * self.n / 3)) - 1


def cordaro_wagner_generator_matrix(n: int) -> npt.NDArray[np.integer]:
    # See [CW67].
    d = int(np.ceil(2 * n / 3)) - 1
    q = (n + 1) // 3
    return np.hstack((
        np.repeat([[1], [0]], repeats=d - q, axis=1),
        np.repeat([[0], [1]], repeats=q, axis=1),
        np.repeat([[1], [1]], repeats=n - d, axis=1),
    ))



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from . import base


@dataclass
class ManchesterPulse(base.Pulse):
    r"""
    Manchester pulse. It is a [pulse](/ref/Pulse) with waveform given by
    $$
        h(t) =
        \begin{cases}
            -1, & 0 \leq t <  1/2, \\\\
            1, & 1/2 \leq t < 1, \\\\
            0, & \text{otherwise},
        \end{cases}
    $$
    and spectrum given by
    $$
        \hat{h}(f) = \operatorname{sinc}^2 \left( \frac{f}{2} \right) \, \sin^2 \left( \frac{\pi f}{2} \right).
    $$

    The Manchester pulse is depicted below.

    <figure markdown>
      ![Manchester pulse.](/figures/pulse_manchester.svg)
    </figure>

    **Attributes:**

    <span style="font-size: 90%; font-style: italic; color: gray; margin-left: 1em;">(No attributes)</span>

    Examples:
        >>> pulse = komm.ManchesterPulse()
        >>> pulse.waveform([-0.50, -0.25,  0.00,  0.25,  0.50,  0.75,  1.00])
        array([ 0.,  0., -1., -1.,  1.,  1.,  0.])
        >>> pulse.spectrum([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75]).round(4)
        array([0.5249, 0.4053, 0.1391, 0.    , 0.1391, 0.4053, 0.5249])
    """

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        t = np.asarray(t)
        return -1.0 * (0 <= t) * (t < 0.5) + 1.0 * (0.5 <= t) * (t < 1.0)

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        f = np.asarray(f)
        return np.sinc(f / 2) ** 2 * np.sin(np.pi * f / 2) ** 2

    @cached_property
    def support(self) -> tuple[float, float]:
        return (0.0, 1.0)



from collections.abc import Callable

import numpy as np
import numpy.typing as npt

from . import base


def mean_squared_quantization_error(
    quantizer: base.ScalarQuantizer,
    input_pdf: Callable[[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
    input_range: tuple[float, float],
    points_per_interval: int,
) -> float:
    # See [Say06, eq. (9.3)].
    x_min, x_max = input_range
    thresholds = np.concatenate(([x_min], quantizer.thresholds, [x_max]))
    mse = 0.0
    for i, level in enumerate(quantizer.levels):
        left, right = thresholds[i], thresholds[i + 1]
        x = np.linspace(left, right, num=points_per_interval, dtype=float)
        pdf = input_pdf(x)
        integrand: npt.NDArray[np.floating] = (x - level) ** 2 * pdf
        integral = np.trapezoid(integrand, x)
        mse += float(integral)
    return mse



from .BCJRDecoder import BCJRDecoder
from .BerlekampDecoder import BerlekampDecoder
from .ExhaustiveSearchDecoder import ExhaustiveSearchDecoder
from .ReedDecoder import ReedDecoder
from .SyndromeTableDecoder import SyndromeTableDecoder
from .ViterbiDecoder import ViterbiDecoder
from .WagnerDecoder import WagnerDecoder

__all__ = [
    "BCJRDecoder",
    "BerlekampDecoder",
    "ExhaustiveSearchDecoder",
    "ReedDecoder",
    "SyndromeTableDecoder",
    "ViterbiDecoder",
    "WagnerDecoder",
]



from dataclasses import dataclass
from functools import cache
from math import comb

import numpy as np
import numpy.typing as npt

from .._util.docs import mkdocstrings
from .BlockCode import BlockCode


@mkdocstrings(filters=["!.*"])
@dataclass(eq=False)
class SingleParityCheckCode(BlockCode):
    r"""
    Single parity-check code. For a given length $n \geq 1$, it is the [linear block code](/ref/BlockCode) whose codewords are obtained by extending $n - 1$ information bits with a single parity-check bit. The repetition code has the following parameters:

    - Length: $n$.
    - Dimension: $k = n - 1$.
    - Redundancy: $m = 1$.
    - Minimum distance: $d = 2$.

    Notes:
        - Its dual is the [repetition code](/ref/RepetitionCode).

    Attributes:
        n (int): The length $n$ of the code. Must be a positive integer.

    Examples:
        >>> code = komm.SingleParityCheckCode(5)
        >>> (code.length, code.dimension, code.redundancy)
        (5, 4, 1)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1],
               [0, 1, 0, 0, 1],
               [0, 0, 1, 0, 1],
               [0, 0, 0, 1, 1]])
        >>> code.check_matrix
        array([[1, 1, 1, 1, 1]])
        >>> code.minimum_distance()
        2

        >>> code = komm.SingleParityCheckCode(16)
        >>> code.codeword_weight_distribution()
        array([    1,     0,   120,     0,  1820,     0,  8008,     0, 12870,
                   0,  8008,     0,  1820,     0,   120,     0,     1])
        >>> code.coset_leader_weight_distribution()
        array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    """

    n: int

    def __post_init__(self):
        if not self.n >= 1:
            raise ValueError("n must be a positive integer")
        super().__init__(check_matrix=np.ones((1, self.n), dtype=int))

    @cache
    def minimum_distance(self) -> int:
        return 2

    @cache
    def codeword_weight_distribution(self) -> npt.NDArray[np.integer]:
        n = self.n
        codeword_weight_distribution = np.zeros(n + 1, dtype=int)
        for w in range(0, n + 1, 2):
            codeword_weight_distribution[w] = comb(n, w)
        return codeword_weight_distribution
